/* Bit vector set operations */

#include "main.h"

#define BITSINLONG 32
#define BYTESINLONG 4

bvs
*bvs_create(int num_bits, int max_num_vectors)
{
bvs *bvsp;
int mem_required;

	mem_required = BYTESINLONG * ((num_bits / BITSINLONG) + 1)
				   * max_num_vectors;
	bvsp = (bvs *) memory (sizeof(bvs),"Create bvs structure");
	bvsp->data = (unsigned long (*)[]) memory(mem_required,"Error - Creating Bit Vector Set");
	bvsp->vector_size = num_bits;
	bvsp->vector_long_size = ((num_bits / BITSINLONG) + 1);
	bvsp->end_bits = num_bits % BITSINLONG;
	bvsp->max_vectors = max_num_vectors;
	return (bvsp);
}

int bvs_bit_count_table[] =
  { 0, 1, 1, 2, 1, 2, 2, 3, 
    1, 2, 2, 3, 2, 3, 3, 4, 
    1, 2, 2, 3, 2, 3, 3, 4, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    1, 2, 2, 3, 2, 3, 3, 4, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    1, 2, 2, 3, 2, 3, 3, 4, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    4, 5, 5, 6, 5, 6, 6, 7, 
    1, 2, 2, 3, 2, 3, 3, 4, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    4, 5, 5, 6, 5, 6, 6, 7, 
    2, 3, 3, 4, 3, 4, 4, 5, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    4, 5, 5, 6, 5, 6, 6, 7, 
    3, 4, 4, 5, 4, 5, 5, 6, 
    4, 5, 5, 6, 5, 6, 6, 7, 
    4, 5, 5, 6, 5, 6, 6, 7, 
    5, 6, 6, 7, 6, 7, 7, 8 };


int
bvs_count_bits (bvs *bvsp, int v)
{
  int count = 0;
  unsigned long t;
  int i;
  int k;

  for(i = 0; i < bvsp->vector_long_size; i++) {
    t = (*(bvsp->data))[bvsp->vector_long_size*v + i];
    if (i == bvsp->vector_long_size-1) {
      for(k = 0; k < bvsp->end_bits; k++) {
	if (t & ((unsigned long) 1 << k)) count++;
      }
    } else {
      count += bvs_bit_count_table[((unsigned long) 255) & t];
      count += bvs_bit_count_table[((unsigned long) 255) & (t >> 8)];
      count += bvs_bit_count_table[((unsigned long) 255) & (t >> 16)];
      count += bvs_bit_count_table[((unsigned long) 255) & (t >> 24)];
    }
  }
  return (count);
}


void
bvs_intersection (bvs *bvsp, int v1, int v2, int vout)
{
int vs = bvsp->vector_long_size;
int i;

	for(i = 0; i < bvsp->vector_long_size; i++) {
		(*bvsp->data)[vs*vout+i] = (*bvsp->data)[vs*v1+i] &
								   (*bvsp->data)[vs*v2+i];
	}
}

void
bvs_and (bvs *bvsp, int v1, int v2, int vout)
{
int vs = bvsp->vector_long_size;
int i;

	for(i = 0; i < bvsp->vector_long_size; i++) {
		(*bvsp->data)[vs*vout+i] = (*bvsp->data)[vs*v1+i] &
								   (*bvsp->data)[vs*v2+i];
	}
}

void
bvs_or (bvs *bvsp, int v1, int v2, int vout)
{
int vs = bvsp->vector_long_size;
int i;

	for(i = 0; i < bvsp->vector_long_size; i++) {
		(*bvsp->data)[vs*vout+i] = (*bvsp->data)[vs*v1+i] |
								   (*bvsp->data)[vs*v2+i];
	}
}

void
bvs_copy (bvs *bvsp, int v1, int vout)
{
int vs = bvsp->vector_long_size;
int i;

	for(i = 0; i < bvsp->vector_long_size; i++) {
		(*bvsp->data)[vs*vout+i] = (*bvsp->data)[vs*v1+i];
	}
}

void
bvs_clear (bvs *bvsp, int v, int val)
{
int i;

	unsigned long m = val ? (unsigned long) 0xFFFFFFFF :
							(unsigned long) 0x00000000 ;
	for(i = 0; i < bvsp->vector_long_size; i++) {
		(*bvsp->data)[bvsp->vector_long_size*v + i] = m;
	}
}

void
bvs_set (bvs *bvsp, int v, int bit, int val)
{
  int vs = (bvsp->vector_long_size*v) + (bit / BITSINLONG);
  if (val) {
    (*bvsp->data)[vs] = (*bvsp->data)[vs] |
      ((unsigned long) 0x00000001 << (bit % BITSINLONG));
  } else {
    (*bvsp->data)[vs] = (*bvsp->data)[vs] &
      ~((unsigned long) 0x00000001 << (bit % BITSINLONG));
  }
}



int
bvs_get (bvs *bvsp, int v, int bit)
{
  register int vs = (bvsp->vector_long_size*v) + (bit / BITSINLONG);
  register int r = bit % BITSINLONG;
  
  return (((*bvsp->data)[vs] & (((unsigned long) 0x00000001) << r)) >> r);
}

void
bvs_show (bvs *bvsp)
{
  int i,j;
  
  for (i=0;i<bvsp->vector_size;i++) {
    for (j=0;j<bvsp->max_vectors;j++) {
      if (bvs_get(bvsp,j,i )) {
	printf("1");
      } else {
	printf("0");
      }
		}
    printf ("\n");
  }
}

void
bvs_show_index (lst index, bvs *bvsp)
{
  int i,j,size,v,element;
  
  size = lst_size(index);
  for (j=0;j<size;j++) printf(" %1d",(j/10)%10); printf("\n");
  for (j=0;j<size;j++) printf(" %1d",j%10); printf("\n");

  for (i=0;i<bvsp->vector_size;i++) {
    for (j=0;j<size;j++) {
      element = lst_look(index,j);
      if (element == -1) {
	printf(" X");
      } else {
	v = bvs_get(bvsp,element,i);
	if (v) {
	  printf(" 1");
	} else {
	  printf(" 0");
	}
      }
    }
    printf ("\n");
  }
}


float
bvs_dot_f (bvs *bvsp, int vec, float fv[])
{
  int i;
  float dot;

  dot = 0.0;
  for (i=0;i<bvsp->vector_size;i++) {
      if (bvs_get(bvsp,vec,i)) {
	dot += fv[i];
      }
  }
  return (dot);
  /* int i;
  float sum;
  register unsigned long t;
  int vs = bvsp->vector_long_size;

  sum = 0.0;
  
  if (bvsp->vector_size != 0) {
    t = (*bvsp->data)[vs*vec+(0/BITSINLONG)];
    if (t & ((unsigned long) 0x00000001)) sum += fv[0];
    t = t >> 1;
    for (i=1;i<bvsp->vector_size;i++) {
      if ((i % BITSINLONG) == 0) t = (*bvsp->data)[vs*vec+(i/BITSINLONG)];
      if (t & ((unsigned long) 0x00000001)) sum += fv[i];
      t = t >> 1;
    }
  }
  return(sum);*/
}

void
bvs_Union(bvs *b, int num, int list[], int out)
{
int i,j;
unsigned long t;

	for (i=0;i<b->vector_long_size;i++) {
		t = (unsigned long) 0x00000000;
		for (j=0;j<num;j++) {
			t |= (*b->data)[b->vector_long_size*list[j] + i];
		}
		(*b->data)[b->vector_long_size*out + i] = t;
	}
}

int
bvs_hash_fun(bvs *b, int pos)
{
  unsigned long t;
  int i, k;
  unsigned long c;

  c = 0;
  for(i = 0; i < b->vector_long_size; i++) {
    t = (*(b->data))[b->vector_long_size*pos + i];
    if (i == b->vector_long_size-1) {
      for(k = 0; k < b->end_bits; k++) {
	if (t & ((unsigned long) 1 << k)) c = (c + look_rlist(i+k)) % (unsigned long) 57123;
      }
    } else {
      c = (c + look_rlist(i)*t) % (unsigned long) 57123;
    }
  }
  return (c % MAXINT);
}

int
bvs_leq(bvs *bvsp, int v1, int v2) /* is v1 <= v2 */
{	   
unsigned long t1, t2;
int i;
int k;

	for(i = 0; i < bvsp->vector_long_size; i++) {
		t1 = (*(bvsp->data))[bvsp->vector_long_size*v1 + i];
		t2 = (*(bvsp->data))[bvsp->vector_long_size*v2 + i];
		if (i == bvsp->vector_long_size-1) {
			for(k = 0; k < bvsp->end_bits; k++) {
				if ( (t1 & ((unsigned long) 1 << k)) && 
                                    !(t2 & ((unsigned long) 1 << k))) return (FALSE);
			}
		} else {
		  if (t1 & (~t2)) return (FALSE);
		}
	}
	return (TRUE);
}

int
bvs_eq(bvs *bvsp, int v1, int v2) /* is v1 == v2 */
{	   
unsigned long t1, t2;
int i;
int k;

	for(i = 0; i < bvsp->vector_long_size; i++) {
		t1 = (*(bvsp->data))[bvsp->vector_long_size*v1 + i];
		t2 = (*(bvsp->data))[bvsp->vector_long_size*v2 + i];
		if (i == bvsp->vector_long_size-1) {
			for(k = 0; k < bvsp->end_bits; k++) {
				if ( (t1 & ((unsigned long) 1 << k)) != 
                                     (t2 & ((unsigned long) 1 << k))) return (FALSE);
			}
		} else {
		  if (t1 != t2) return (FALSE);
		}
	}
	return (TRUE);
}


void
bvs_test ()
{
	bvs *b;


	b = bvs_create(34,10);
	bvs_clear (b,0,0);
	bvs_clear (b,1,1);
	bvs_clear (b,2,0);
	bvs_clear (b,3,1);
	printf("First test \n");
	bvs_show (b);
	printf("\nSecond test\n");
	bvs_set (b,2,3,1);
	bvs_set (b,3,33,0);
	bvs_set (b,0,32,1);
	bvs_set (b,1,31,0);
	bvs_show (b);
	if (bvs_get(b,0,32)) printf ("okay get test1\n");
	printf ("0 count %d\n",bvs_count_bits (b,0));
	printf ("1 count %d\n",bvs_count_bits (b,1));
	printf ("2 count %d\n",bvs_count_bits (b,2));
	printf ("3 count %d\n",bvs_count_bits (b,3));
}

