/* boost -  implementing boosting and bagging for lime */

#include "main.h"

float (*init_weight)[];

	  
float
error_cal (float pc, float np, float nc, float nn) {
  return (((np-pc)+nc)/(np + nn));
}

void
boost_test(pred_set *ps, boost *bc,int *c,int *n, info *i) {
  int k, j, arity;
  int sum;
  lst data;
  float ws;

  sum = 0;
  if (ps->num != 0) {
    arity = ps->preds[0].arity;
    data = lst_new(arity);
    for (k=0;k<(ps_count(ps));k++) {
      lst_copy(lst_look(ps->preds[0].preds,k),data);
      ws = 0.0;
      for(j=0;j<bc->num_trials;j++) {
	if (cover_ex2(i,lst_look(bc->programs,j),data)) {
	  ws += (*bc->weight)[j];
	} else {
	  ws -= (*bc->weight)[j];
	}
      }
      if (ws >= 0.0) sum += lst_look(ps->preds[0].number,k);
      if (DB(dbperform)) {
	printf("(%1d)",lst_look(ps->preds[0].number,k));
	show_an_example(data,ps->preds[0].type);
	printf("\n");
      }
    }
    lst_free(data);
  
    *n = lst_sum(ps->preds[0].number);
    *c = sum;
  } else {
    *n = 0;
    *c = 0;
  }
}
	

boost
*boost_newc(int num_trials) {
  boost *b;

  b = (boost *) memory(sizeof(boost),"creating boost classifier");
  b->programs = lst_new(num_trials);
  b->weight = (float (*)[]) memory (sizeof(float)*num_trials,"weights for boost classifier");
  b->num_trials = num_trials;
  return b;
}	



void
boost_normalize_weights(tab *t)
{  /* assumes num_w_ is set correctly */
  int j;
  float scale,sum;

  sum = 0.0;
  for (j=0;j<t->num_e_p;j++) {
    sum += (*t->weight_p)[j];
  }
  scale = t->num_w_p;
  for (j=0;j<t->num_e_p;j++) {
    (*t->weight_p)[j] = ((*t->weight_p)[j]*scale)/sum;
  }
 
  sum = 0.0;
  for (j=0;j<t->num_e_n;j++) {
    sum += (*t->weight_n)[j];
  }
  scale = t->num_w_n;
  for (j=0;j<t->num_e_n;j++) {
    (*t->weight_n)[j] = ((*t->weight_n)[j]*scale)/sum;
  }
 
}

void
boost_reset_search(search *ss, info *i)
{
  lp_reset(ss->clause_best,i->the_tab);
  lp_reset(ss->clause_set_final,i->the_tab);
}

float
cal_error_weight(int lit,tab *t) 
{
  float sum;
  int j;

  sum = 0.0;

  for(j=0;j<t->num_e_p;j++) {
    if (!bvs_get(t->lit_cover_p,lit,j)) {
      sum += (*t->weight_p)[j];
    }       
  }

  for(j=0;j<t->num_e_n;j++) {
    if (bvs_get(t->lit_cover_n,lit,j)) {
      sum += (*t->weight_n)[j];
    }
  }  
  return (sum / (t->num_w_p + t->num_w_n));
}

void
boost_change_weights(tab *t,int lit,float error)
{
  float dm,dc;
  int j;

  dm = 2 * error;
  dc = 2 * (1 - error);

  for(j=0;j<t->num_e_p;j++) {
    if (bvs_get(t->lit_cover_p,lit,j)) {
      (*t->weight_p)[j] = (*t->weight_p)[j]/dc;
    } else {
      (*t->weight_p)[j] = (*t->weight_p)[j]/dm;
    }
  }

  for(j=0;j<t->num_e_n;j++) {
    if (bvs_get(t->lit_cover_n,lit,j)) {
      (*t->weight_n)[j] = (*t->weight_n)[j]/dm;
    } else {
      (*t->weight_n)[j] = (*t->weight_n)[j]/dc;
    }
  }
}


void
boosth_change_weights(tab *t,int lit,float error, int boosthval)
{
  float dm,dc,nw;
  int j;

  dm = 2 * error;
  dc = 2 * (1 - error);

  for(j=0;j<t->num_e_p;j++) {
    if (bvs_get(t->lit_cover_p,lit,j)) {
      (*t->weight_p)[j] = (*t->weight_p)[j]/dc;
    } else {
      nw   = (*t->weight_p)[j]/dm;
      if (nw >= (2.0 * (*init_weight)[j])) {
	(*t->weight_p)[j] = (2.0 * (*init_weight)[j]);
      } else {
	(*t->weight_p)[j] = nw;
      }
    }
  }

  for(j=0;j<t->num_e_n;j++) {
    if (bvs_get(t->lit_cover_n,lit,j)) {
      nw   = (*t->weight_n)[j]/dm;
      if (nw >= (2.0 * (*init_weight)[j+t->num_e_p])) {
	(*t->weight_n)[j] = (2.0 * (*init_weight)[j+t->num_e_p]);
      } else {
	(*t->weight_n)[j] = nw;
      }
    } else {
      (*t->weight_n)[j] = (*t->weight_n)[j]/dc;
    }
  }
  boost_normalize_weights(t);
  
}

void bag_reweight (tab *t) {
  int j,k, found;
  float r, num_w, sum;

  num_w = t->num_w_p+t->num_w_n;

  for(j=0;j<t->num_e_p;j++) {
    (*t->weight_p)[j] = 0.0;
  }
  
  for(j=0;j<t->num_e_n;j++) {
    (*t->weight_n)[j] = 0.0;
  }

  for (k=0;k<num_w;k++) {
    /*r = my_random_float(100.0); printf("%3.3f %3.3f %3.3f %3.3f \n",r,num_w,t->num_w_p,t->num_w_n);*/
    r = my_random_float(num_w);
    /*printf("%3.3f %3.3f %3.3f %3.3f \n",r,num_w,t->num_w_p,t->num_w_n);*/
    if (r < t->num_w_p) {
      sum = 0.0;
      found = 0;
      for(j=0;j<t->num_e_p && !found ;j++) {
	sum += (*init_weight)[j];
	if (r <= sum) {
	  found = 1;
	  (*t->weight_p)[j] += 1.0;
	}
      } 
     
    } else {
      sum = t->num_w_p;
      found = 0;
      for(j=0;j<t->num_e_n && !found ;j++) {
	sum += (*init_weight)[j+t->num_e_p];
	if (r <= sum) {
	  found = 1;
	  (*t->weight_n)[j] += 1.0;
	}
      } 
    }
  }
}

void init_bag (tab *t) {
  int j;

  init_weight = (float (*)[]) memory (sizeof(float)*(t->num_e_p+t->num_e_n), "weights for bag classifier");
  for(j=0;j<t->num_e_p;j++) {
    (*init_weight)[j] = (*t->weight_p)[j];
  } 
  for(j=0;j<t->num_e_n;j++) {
    (*init_weight)[j+t->num_e_p] = (*t->weight_n)[j];
  } 
}

void 
boost_do (info *i,search *ss, boost *bc, int num_trials) 
{
  int trial;
  tab *t;
  float error;
  int done;
  int lit;
  lst tp, p;

  /* set weight to 1/n for each element */
  t = i->the_tab;
  /*  boost_normalize_weights(t); */

  bc->num_trials = 0;
  done = FALSE;
  for(trial=0;trial<num_trials && !done; trial++) {
    if (DB(dbstage)) printf ("= boosting trial %2d\n",trial);
    do_search(i,ss);
    lit = lp_bestlit(ss->clause_set_final);
    error = cal_error_weight(lit,t);
    tp = lp_bestlp(ss->clause_set_final);

    p = lst_new(lst_size(tp));
    lp_copy(tp,p);

    if (error == 0.0) {
      done = TRUE;
      bc->num_trials = 1;
      (*bc->weight)[0] = 1.0;
      lp_tree_remove(bc->programs,4);
      bc->programs = lst_new(1);
      lst_append(bc->programs,p);
    } else if (error >= 0.5) {
      done = TRUE;
      lp_tree_remove(p,3);
    } else {
      (*bc->weight)[trial] = log ((1-error)/error);
      lst_append(bc->programs,p);
      bc->num_trials++;
      boost_change_weights(t,lit,error);
    }
    if (DB(dbboost)) {
      printf("== bosting error %6.3f ",error);
      if (done) {
	printf("boosing done\n");
      } else {
	printf("weight %6.3f\n",(*bc->weight)[trial]);
      }
      show_lp_test (stdout, tp, i);
    }
    boost_reset_search(ss,i);
  }
}  



void 
boosth_do (info *i,search *ss, boost *bc, int num_trials, int boosthval) 
{
  int trial;
  tab *t;
  float error;
  int done;
  int lit;
  lst tp, p;

  /* set weight to 1/n for each element */
  t = i->the_tab;
  /*  boost_normalize_weights(t); */

  init_bag(t);

  bc->num_trials = 0;
  done = FALSE;
  for(trial=0;trial<num_trials && !done; trial++) {
    if (DB(dbstage)) printf ("= boostingh %d trial %2d\n",trial,boosthval);
    do_search(i,ss);
    lit = lp_bestlit(ss->clause_set_final);
    error = cal_error_weight(lit,t);
    tp = lp_bestlp(ss->clause_set_final);

    p = lst_new(lst_size(tp));
    lp_copy(tp,p);

    if (error == 0.0) {
      done = TRUE;
      bc->num_trials = 1;
      (*bc->weight)[0] = 1.0;
      lp_tree_remove(bc->programs,4);
      bc->programs = lst_new(1);
      lst_append(bc->programs,p);
    } else if (error >= 0.5) {
      done = TRUE;
      lp_tree_remove(p,3);
    } else {
      (*bc->weight)[trial] = log ((1-error)/error);
      lst_append(bc->programs,p);
      bc->num_trials++;
      boosth_change_weights(t,lit,error,boosthval);
    }
    if (DB(dbboost)) {
      printf("== bostingh %d error %6.3f ",boosthval,error);
      if (done) {
	printf("boosingh %d done\n",boosthval);
      } else {
	printf("weight %6.3f\n",(*bc->weight)[trial]);
      }
      show_lp_test (stdout, tp, i);
    }
    boost_reset_search(ss,i);
  }
}  


void
boost_show (boost *bc, info *i) {
  int j;

  for (j=0;j<bc->num_trials;j++) {
    printf("t = %3d   weight = %6.3f\n",j,(*bc->weight)[j]);
    show_lp_test (stdout, lst_look(bc->programs,j), i);
  }

}



void 
bag_do (info *i,search *ss, boost *bc, int num_trials) 
{
  int trial;
  tab *t;
  float error;
  int done;
  int lit;
  lst tp, p;

  /* set weight to 1 for each element */
  t = i->the_tab;
  /*  boost_normalize_weights(t); */

  init_bag(t);

  bc->num_trials = 0;
  done = FALSE;
  for(trial=0;trial<num_trials && !done; trial++) {
    if (DB(dbstage)) printf ("= bagging trial %2d\n",trial);
    bag_reweight(t);
    /*for (k=0;k<t->num_e_n;k++) printf ("%5e  ",(*t->weight_n)[k]);*/
    do_search(i,ss);
    lit = lp_bestlit(ss->clause_set_final);
    error = cal_error_weight(lit,t);
    tp = lp_bestlp(ss->clause_set_final);

    p = lst_new(lst_size(tp));
    lp_copy(tp,p);

  
    (*bc->weight)[trial] = 1.0; /*log ((1-error)/error);*/
    lst_append(bc->programs,p);
    bc->num_trials++;
    
    if (DB(dbboost)) {
      printf("== bagging error %6.3f ",error);
      if (done) {
	printf("bagging done\n");
      } else {
	printf("weight %6.3f\n",(*bc->weight)[trial]);
      }
      show_lp_test (stdout, tp, i);
    }
    boost_reset_search(ss,i);
  }
}  








