/* input.c */

#include <stdlib.h>
#include<string.h>
#include"main.h"

#define BSIZE 255

 
FILE *infile;
int lineno;
int tokenval;
char lexbuf[BSIZE];

void
init(char file_name[])
{
	lineno = 1;
	if ((infile = fopen(file_name,"r")) == NULL)
	  error("cant open input file");
}

token
lexan()
{
	int t;
	/* long pos; */
	/*	float fval;*/
	int ival;

	while (1) {
		t = fgetc(infile);
		if (t == ' ' || t == '\t')
			;
		else if (t == '\n')
			lineno++;
		else if (t == ',')
			return COMMA;
		else if (t == '.')
			return DOT;
		else if (t == '(')
		  return OPEN_B;
		else if (t == ')')
		  return CLOSE_B;
		else if (t == '\%') {
		  while (t != '\n' && t != EOF) {
		    t = fgetc(infile);
		  }
		  lineno++;
		}
		else if (isdigit(t)) {
			ungetc(t,infile);
/*			pos = ftell(infile);
			if (fscanf(infile,"%e",&fval) == 1) {
				tokenval = (int) fval;
				return REAL;
			} else {
				fseek(infile,pos,SEEK_SET); */
				if (fscanf(infile,"%d",&ival) == 1) {
					tokenval = ival;
					return NUM;
				} else {
					l_error("lex - scaning numbers");
				}
			/*}*/
		}
		else if (t == '\'') {
			int p, b = 0;
			while ((b == 0) || (t != '\'')) {
				lexbuf[b] = t;
				t = fgetc(infile);
				b++;
				if (b >= BSIZE) l_error ("compiler error");
			}
			lexbuf[b] = '\'';
			b++;
			if (b >= BSIZE) l_error ("compiler error");
			lexbuf[b] = '\0';
			
			p = lookup(lexbuf);
			if (p == -1) p = insert(lexbuf);
			tokenval = p;
			return STR;		
		}
		else if (isalpha(t) || t == '[' || t == ']') {
			int p, b = 0;
			while (isalnum(t) || (t == '+') || (t == '-') || (t == '_') || 
			       (t == '[') || (t == ']')) {
				lexbuf[b] = t;
				t = fgetc(infile);
				b++;
				if (b >= BSIZE) l_error ("compiler error");
			}
			lexbuf[b] = '\0';
			if (t != EOF) ungetc(t,infile);
			if (strcmp(lexbuf,"E+") == 0) return E_POS;
			if (strcmp(lexbuf,"E-") == 0) return E_NEG;
			if (strcmp(lexbuf,"T+") == 0) return T_POS;
			if (strcmp(lexbuf,"T-") == 0) return T_NEG;
			if (strcmp(lexbuf,"BG") == 0) return BG;

			p = lookup(lexbuf);
			if (p == -1) p = insert(lexbuf);
			tokenval = p;
			return STR;
		}
		else if (t == EOF) {
			return DONE;
		}
		else
		{
			l_error ("unknown token");
		}
	}
}

void
show_token (token t)
{
  switch (t) 
    {
    case DONE : printf("DONE\n"); break;
    case NONE : printf("NONE\n"); break;
    case E_POS : printf("E_POS\n"); break;
    case E_NEG : printf("E_NEG\n"); break;
    case BG : printf("BG\n"); break;
    case NUM : printf("NUM %d ",tokenval); break;
    case REAL : printf("REAL"); break;
    case STR : printf("STR %d ", tokenval); break;
    case COMMA : printf("COMMA "); break;
    case DOT : printf("DOT\n"); break;
    case OPEN_B : printf("OPEN_B "); break;
    case CLOSE_B : printf("CLOSE_b"); break;
    case T_POS : printf("T_POS\n"); break; 
    case T_NEG : printf("T_NEG\n"); break; 
    }
}

void
l_error (char m[])
{
	fprintf(stderr, "Error line %d: %s\n", lineno, m);
	exit(1);
}


/* the symbol table */

#define INITSTRMAX 100000
#define INITSYMMAX  200000

char (*lexemes)[];
int lastchar = -1;
entrytd (*symtable)[];
int numentry = 0;
lst symtmap;
int STRMAX;
int SYMMAX;

void
init_symtab() {
  int j;

  STRMAX = INITSTRMAX;
  SYMMAX = INITSYMMAX;
  
  lexemes = memory(STRMAX,"allocating memory for lexemes"); 
  symtable = memory(SYMMAX*sizeof(entrytd),"allocating memory for symtable"); 

  symtmap = lst_new(SYMMAX+2);
  for (j=0;j<SYMMAX;j++)
    (*symtable)[j].lexptr = NULL;

}

int
sym_hash(char s[])
{
  int t, j;

  t = 0;
  for (j=0;j<strlen(s);j++) {
    t += look_rlist(j)*((int) s[j]) % 60000;
  }
  return (t % SYMMAX);
}

int
lookup(char s[])
{
	int hp, j;

	hp = sym_hash(s);
	j = hp;
	while ((*symtable)[j].lexptr != NULL) {
	  if (strcmp((*symtable)[j].lexptr,s) == 0) return (*symtable)[j].lexnum;
	  j = (j + 1) % SYMMAX;
	  if (j == hp) return -1;
	}
	return -1;
}


void
expandlexems() {
  char (*newlexemes)[];
  int i,len;

  if (DB(dbstage)) printf("===expanding lexemes  %d -> %d \n",STRMAX,STRMAX*2);

  STRMAX = STRMAX*2;
  newlexemes = memory(STRMAX,"allocating memory for lexemes"); 

  lastchar = -1;
  for (i=0;i<SYMMAX;i++) {
    if ((*symtable)[i].lexptr != NULL) {
    len = strlen((*symtable)[i].lexptr);
    strcpy (&(*newlexemes)[lastchar+1],(*symtable)[i].lexptr);
    (*symtable)[i].lexptr = &(*newlexemes)[lastchar+1];
    lastchar += len + 1;
    }
  }
  mem_free(lexemes);
  lexemes = newlexemes;

  if (DB(dbstage)) printf("===done expanding \n");
}

int
insert (char s[])
{
	int len, j, hp, p;

	len = strlen(s);
	if (numentry + 1 >= SYMMAX) l_error ("entry array full");
	if (lastchar + len + 1 >= STRMAX)  expandlexems();
	/*l_error ("lexemes array full");*/
	hp = sym_hash(s);
	j = hp;
	while ((*symtable)[j].lexptr != NULL) {
	  j = (j + 1) % SYMMAX;
	}
	(*symtable)[j].lexptr = &(*lexemes)[lastchar+1];
	lastchar += len + 1;
	strcpy ((*symtable)[j].lexptr, s);
	p = numentry;
	(*symtable)[j].lexnum = p;
	lst_append(symtmap,j);
	numentry++;
	return p;
}

/* parser */

int lookahead;

void
parse()
{
	lookahead = lexan();
	p_epos();
	if (lookahead == E_NEG) p_eneg();
	if (lookahead == BG) p_bg();
	have_tests = FALSE;
	if (lookahead == T_POS) p_tpos();
	if (lookahead == T_NEG) p_tneg();

}

void
p_epos()
{
	match(E_POS);
	p_perd_list(ps_epos);
}

void
p_eneg()
{
	match(E_NEG);
	p_perd_list(ps_eneg);
}
void
p_tpos()
{
	have_tests = TRUE;
	match(T_POS);
	p_perd_list(ps_tpos);
}

void
p_tneg()
{
	have_tests = TRUE;
	match(T_NEG);
	p_perd_list(ps_tneg);
}

void
p_bg()
{
	match(BG);
	p_perd_list(ps_bg);
}


void
p_perd_list(pred_set *ps)
{
	while (lookahead == STR)
		p_perd(ps);
}

void
p_perd(pred_set *ps)
{
	int name, i, j, tval[MAXARITY];
	token ttype[MAXARITY];
	lst tp, d;

	name = match(STR);
	match(OPEN_B);
	i = 0;
	while (lookahead != CLOSE_B) {
		p_term(&ttype[i],&tval[i]);
		i++;
		if (lookahead == COMMA) match(COMMA);
	}
	match(CLOSE_B);
	match(DOT);
	tp = lst_new(i);
	d = lst_new(i);
	for (j=0;j<i;j++) {
		switch (ttype[j]) {
		   case REAL : lst_append(tp,REAL_TYPE); break;
		   case NUM : lst_append(tp,INT_TYPE); break;
		   case STR : lst_append(tp,SS_TYPE); break;
		   default : l_error ("matching REAL, NUM, or STR"); break; 
		}
		lst_append(d,tval[j]);
	}
	ps_add(ps,name,i,tp,d);
	lst_free(tp);
	lst_free(d);
}

void
p_term(token *t, int *v)
{
	if ((lookahead == NUM) || (lookahead == REAL) || (lookahead == STR)) {
		*t = lookahead;
		*v = tokenval;
		lookahead = lexan();
	} else l_error("expected term");
}

int
match(token t)
{
	int v;
	char e[50];

	v = tokenval;
	if (lookahead == t) {
		lookahead = lexan();
		return (v);
	} else {
		sprintf(e,"token mismatch should be %d was %d",t,lookahead);
		l_error (e);
		return -1;
	}
}

void
show_lp_lst (FILE *f, lp_lst *lpl, info *i)
{
	int j;

	for (j=0;j<lst_size(lpl->index);j++) {
		fprintf(f,"%5.2f, %5.2f\n",DIX(lpl,j,v1),DIX(lpl,j,v2));
		show_lp (f, DIX(lpl,j,lp),i);
	}
}
void

show_lp_lst_test (FILE *f, lp_lst *lpl, info *i)
{
  int j;
  
  for (j=0;j<lst_size(lpl->index);j++) {
    if (DB(dbsearchb)) {
      fprintf(f,"v1 = %5.2f, v2 = %5.2f\n",DIX(lpl,j,v1),DIX(lpl,j,v2));
    }		
    show_lp_test (f, DIX(lpl,j,lp),i);
  }
}


void
show_tab_lp_lst (FILE *f, lp_lst *lpl, info *i)
{
  if (DB(dbsearchb)) {
    fprintf(f,"v1 = %5.2f\t v2 = %5.2f\t",DIX(lpl,0,v1),DIX(lpl,0,v2));
  }
  show_lp_nnl (f, DIX(lpl,0,lp),i);
}

int
count_preds(pred_set *ps)
{
  if (ps->num == 1) {
    return lst_sum(ps->preds[0].number);
  } else if (ps->num == 0) {
    return 0;
  } else {
    error("Only one predicate name allowed in the pos and neg examples");
    return -1;
  }
}

void
show_lp (FILE *f, lst lp, info *i)
{
	int j, k, step, tpos;
	int np,ep,nn,en;
        float pe, bits, r, theta, num_cover;
	tab *t;

	t = i->the_tab;
	/* do tests on data and show results */
	np = count_preds(ps_epos);
 	nn = count_preds(ps_eneg);
        ep = calculate_cover(i,lp,ps_epos);
        en = calculate_cover(i,lp,ps_eneg);
        pe = ((float) (np-ep+en))/((float) (np+nn));

	num_cover = 0.0;                         /* calculate theta cover */
	for (k = 0; k < t->num_theta; k++) {
	  tpos = t->num_e_n + t->num_e_p + k;
	  if (cover_ex(i,lp,&step,tpos)) {
	    num_cover += (*t->weight_t)[k];
	  } 
	}
	theta = cal_theta(t,num_cover);
	bits = prob_lp(lp,i);
	r = bay(i->the_tab,bits,ep,en,theta);

       fprintf(f,"==Data # Pos %d/%d, # Neg %d/%d # Error %5.3f # prob %5.3f # theta %5.3f # q %5.3f\n",ep,np,en,nn,pe,bits,theta,r);

	for (j=0;j<lst_size(lp);j++) {
		show_c (f,lst_look(lp,j),i);
		fprintf (f,"\n");
	}
}

void
show_lp_nostats(FILE *f, lst lp, info *i)
{
	int j;

	for (j=0;j<lst_size(lp);j++) {
		show_c (f,lst_look(lp,j),i);
		fprintf(f,"\n");
	}
}

void
show_lp_test (FILE *f, lst lp, info *i)
{
	int j,  np,ep,nn,en;
        float pe;

	/* do tests on data and show results */
	np = count_preds(ps_epos);
 	nn = count_preds(ps_eneg);
        ep = calculate_cover(i,lp,ps_epos);
        en = calculate_cover(i,lp,ps_eneg);
        pe = ((float) (np-ep+en))/((float) (np+nn));
      
       fprintf(f,"==Data # Pos %d/%d, # Neg %d/%d # Error %5.4f\n",ep,np,en,nn,pe);

	/* do tests and show results */
	np = count_preds(ps_tpos);
 	nn = count_preds(ps_tneg);
        ep = calculate_cover(i,lp,ps_tpos);
        en = calculate_cover(i,lp,ps_tneg);
        pe = ((float) (np-ep+en))/((float) (np+nn));
      
       fprintf(f,"==Tests # Pos %d/%d, # Neg %d/%d # Error %5.4f\n",ep,np,en,nn,pe);


        /* show clause */
	for (j=0;j<lst_size(lp);j++) {
		show_c (f,lst_look(lp,j),i);
		fprintf (f,"\n");
	}
}

void
show_lp_nnl (FILE *f, lst lp, info *i)
{
	int j;

	for (j=0;j<lst_size(lp);j++) {
		show_c (f,lst_look(lp,j),i);
		fprintf(f,"\t");
	}
}

int
all_rec_lits_lp(lst lp, info *i) {
  int j, s;

  s = lst_size(lp);
  for (j=0;j<s;j++) {
   if (no_rec_lits_c(lst_look(lp,j),i)) return FALSE; 
  }
  return TRUE;
}

int
no_rec_lits_lp(lst lp, info *i) {
  int j, s;

  s = lst_size(lp);
  for (j=0;j<s;j++) {
   if (!no_rec_lits_c(lst_look(lp,j),i)) return FALSE; 
  }
  return TRUE;
}

int
no_rec_lits_c(lst c, info *i) {
  int j, s;

  s = lst_size(c);
  for (j=1;j<s;j++) {
   if (!no_rec_lit(lst_look(c,j),i)) return FALSE; 
  }
  return TRUE;
}

int
no_rec_lit(int l, info *i) {
  int lit_num, raw_pred;

  lit_num = (*i->lit_tab)[l].lit_num;
  if (lit_num >= 0) {
    raw_pred = (*i->gen_lit)[lit_num].raw_pred;
    if (raw_pred == -1) {
      return FALSE;
    } else {
      return TRUE;
    }
  } else {
    return FALSE; 
  }
}

void
show_c (FILE *f, lst c, info *i)
{
	int j, s;

	show_l(f,0,i);
	s = lst_size(c);
	if (s > 0) fprintf(f,":-");
	for (j=1;j<s;j++) {
		show_l (f,lst_look(c,j),i);
		if (j+1<lst_size(c)) fprintf(f,",");
	}
	fprintf(f,".");

}

void
show_l (FILE *f, int l, info *i)
{
	int lit_num, j, r;
	int raw_pred;
	lst rib, rob;
	lst inv, outv;
	pred *p;


	lit_num = (*i->lit_tab)[l].lit_num;
	inv = (*i->lit_tab)[l].in_vars;
	outv = (*i->lit_tab)[l].out_vars;

	if (lit_num >= 0) {
		raw_pred = (*i->gen_lit)[lit_num].raw_pred;
		rib = (*i->gen_lit)[lit_num].raw_in_bind;
		rob = (*i->gen_lit)[lit_num].raw_out_bind;
		if (raw_pred == -1) {
		  p = &(ps_epos->preds[0]);
		} else if (raw_pred < 0)
			p = &(ps_invent->preds[-raw_pred - 2]);
		else
			p = &(ps_bg->preds[raw_pred]);
	} else {
		p = &(ps_epos->preds[0]);
		rib = i->inb;
		rob = i->outb;
	}

	fprintf(f,"%s(",(*symtable)[lst_look(symtmap,p->name)].lexptr);

	for (j=0;j<p->arity;j++) {
		if ((r = lst_find(rib,j)) != -1) {
			fprintf(f,"V%d", lst_look(inv,r));
		} else if ((r = lst_find(rob,j)) != -1) {
			fprintf(f,"V%d", lst_look(outv,r));
		} else {
			fprintf(f,"_");
		}
		if (j+1<p->arity) fprintf(f,",");
	}

	fprintf(f,")");

}





