/* invent.c  predicate invention */

#include "main.h"


void  /* map var_table through pos_in_table to table index */ 
trans_vars(info *i, lst v_index, lst vt_index)
{
  int k;

  lst_clear(vt_index);
  for (k=0;k<lst_size(v_index);k++) {
    lst_append(vt_index,((*i->var_tab)[lst_look(v_index,k)]).pos_in_tab);
  }
}

void  /* map var_table through pos_in_table to table index */ 
trans_vars_type(info *i, lst v_index, lst type_index)
{
  int k;

  lst_clear(type_index);
  for (k=0;k<lst_size(v_index);k++) {
    lst_append(type_index,((*i->var_tab)[lst_look(v_index,k)]).type);
  }
}

lst
inv_get_l(tab *t, lst vt_index, int offset, int size)
{
   int k,j;
   lst res, terms;
   int arity;
   
   arity = lst_size(vt_index);
   res = lst_new(size);
   for (k=0;k<size;k++) {
      terms = lst_new(arity);
      for (j=0;j<arity;j++) {
	 lst_append(terms,
		    (*t->var_table)[lst_look(vt_index,j) * t->num_et + 
                                    offset + k]);
      }
      lst_append(res,terms);
   } 
   return res;
}


void /* calculate the invented predicates ground atoms and its coverage */
cal_pred(tab *t, lst vt_index, lst raw_pred, int pos)
{
   int k;
   lst pos_l, neg_l, theta_l, k_pos_l; 

   pos_l   = inv_get_l(t,vt_index,0,                    t->num_e_p);
   neg_l   = inv_get_l(t,vt_index,t->num_e_p,           t->num_e_n);
   theta_l = inv_get_l(t,vt_index,t->num_e_p+t->num_e_n,t->num_theta);
   for (k=0;k<t->num_e_p;k++) {
      k_pos_l =  lst_look(pos_l,k);
      if (lstlst_member(k_pos_l,raw_pred)) {
	 bvs_set(t->lit_cover_p,pos,k,1);
      } else {
	 if (lstlst_member(k_pos_l,neg_l) ||
	     lst_member(-1,k_pos_l)) {
	    bvs_set(t->lit_cover_p,pos,k,0);
	 } else {
	    bvs_set(t->lit_cover_p,pos,k,1);
	    lst_append(raw_pred,lst_replicate(k_pos_l));
	 }
      }
   }
   for (k=0;k<t->num_e_n;k++) {
      bvs_set(t->lit_cover_n,pos,k,0);
   }
   for (k=0;k<t->num_theta;k++) {
      if (lstlst_member(raw_pred,lst_look(theta_l,k))) {
	 bvs_set(t->lit_cover_t,pos,k,1);
      } else {
	 bvs_set(t->lit_cover_t,pos,k,0);
      }
   }
   lp_tree_remove(pos_l,3);
   lp_tree_remove(neg_l,3);
   lp_tree_remove(theta_l,3);
}

int
make_new_pred(info *i, lst v_index, lst type_index, lst lits_used, int pos, lst raw_pred)
{
  static int pred_inv_num = 0;
  char name[20];
  int glit_num, arity, size, tpos, num, k, w, pname;
  int in[MAXARITY], out[MAXARITY];
  lst in_r, out_r;
  map *m;
  int h;

  arity = lst_size(type_index);
  size = lst_size(raw_pred);
  glit_num = i->num_gen_lit;

  /* workout new name */
  num = pred_inv_num;
  sprintf(name,"ip%d",num);
  pname = lookup(name);
  if (pname == -1) pname = insert(name);
  pred_inv_num++;

  /* add to pred set */
  for (k=0;k<size;k++) {
    ps_add(ps_invent,pname,arity,type_index,lst_look(raw_pred,k));
  }

  /* create glit entry */
  if (glit_num == i->max_gen_lit) error("glit is to big - from invent");
  (*i->gen_lit)[glit_num].num_in = arity;
  (*i->gen_lit)[glit_num].num_out = 0;
  (*i->gen_lit)[glit_num].type_in = lst_replicate(type_index);
  (*i->gen_lit)[glit_num].type_out = lst_new(1);
  m = map_create(size, arity, 0);
  for (k=0;k<size;k++) {
    for (w=0;w<arity;w++) {
      in[w] = lst_look(lst_look(raw_pred,k), w);
    }
    map_add_sort(m, in, out, k);
  }
  (*i->gen_lit)[glit_num].fun = m;
  (*i->gen_lit)[glit_num].symtric = FALSE;
  (*i->gen_lit)[glit_num].byjection = FALSE;
  (*i->gen_lit)[glit_num].bits = (3.0 + (2.0 * arity));
  (*i->gen_lit)[glit_num].raw_pred = -2-num;
  in_r = lst_new(arity);
  for (w=0;w<arity;w++)
    lst_append(in_r, w);
  out_r = lst_new(1);
  (*i->gen_lit)[glit_num].raw_in_bind = in_r;
  (*i->gen_lit)[glit_num].raw_out_bind = out_r;
  if (DB(dbinvent)) {
    printf("glit new inv pred %d from %d name %s in ",
	   glit_num,
	   (*i->gen_lit)[glit_num].raw_pred, name);
    lst_show(in_r); printf(" out ");
    lst_show(out_r); printf("\n");
    map_show((*i->gen_lit)[glit_num].fun);
  }
  i->num_gen_lit++;

  /* creat lit tab entry */
  h = tab_bits_hash_fun(i->the_tab,pos); 
  tab_bits_hash_add(i->the_tab, pos, h);
    
  if (i->num_lit_tab == i->max_lit_tab) error ("lit table full - invent pred");
  tpos = i->num_lit_tab;  /* set up the ltab */
  i->num_lit_tab++;  
  (*i->lit_tab)[tpos].lit_num = glit_num;
  (*i->lit_tab)[tpos].level = 10;
  (*i->lit_tab)[tpos].in_vars = lst_new(arity);
  for (k=0;k<arity;k++) 
    lst_append((*i->lit_tab)[tpos].in_vars,lst_look(v_index,k));
  (*i->lit_tab)[tpos].out_vars = lst_new(0);
  (*i->lit_tab)[tpos].lit_pos = pos;
  (*i->lit_tab)[tpos].lits_used = lst_replicate(lits_used);
  lst_append((*i->lit_tab)[tpos].lits_used,glit_num);
  (*i->lit_tab)[tpos].bits = (*i->gen_lit)[glit_num].bits;

  return tpos;
}

lst   
get_lits_used(info *i,lst v_index)
{
  lst r, temp;
  int k;

  r = lst_new(10);
  temp = lst_new(10);
  for (k=0;k<lst_size(v_index);k++) {
    lst_merge(r,(*i->lit_tab)[(*i->var_tab)[lst_look(v_index,k)].from_lit].lits_used,temp);
    lst_copy(temp,r);
  }
  lst_free(temp);
  return r;
}

/* asseses and adds predicate if needed */
void
asses_pred(search *ss,tab *t, info *i, lst v_index)
{
  int pos;
  float pc, nc, tc;
  lst raw_pred,lu,p;
  lst vt_index, type_index;
  float theta, bits, poss_r;
  pos = fl_get(t->lit_free_list);            /* place to cover */
  
  vt_index = lst_new(lst_size(v_index));
  type_index = lst_new(lst_size(v_index));
   
  trans_vars(i,v_index,vt_index);
  trans_vars_type(i,v_index,type_index);
  raw_pred = lst_new(t->num_e_p);
  cal_pred(t,vt_index,raw_pred,pos);
  
  cal_cover(t,pos);
  tab_cover(t,pos,&pc,&nc,&tc);
    
  lu = get_lits_used(i,v_index);
  bits = prob_clause(lu,i)*(lst_size(raw_pred));
  
  theta = cal_theta(t,tc);
  poss_r = best_bay(t,bits,pc,nc,theta) + 2700.0;

  if (DB(dbinvent)) {
    printf("inv- pc %6.2f nc %6.2f tc %6.2f bits %4.2f theta %4.2f poss_r %4.2f lp_worse %4.2f ",pc,nc,tc,bits,theta,poss_r,lp_worse(ss->clause_best));
    printf("invsize : %4d => ",lst_size(raw_pred));
  }
  if (poss_r > lp_worse(ss->clause_best) &&
      (((float) lst_size(raw_pred)) < pc)) {
    if (DB(dbinvent)) printf("Add it\n");
    p = make_new_pred(i,v_index, type_index,lu,pos,raw_pred);   
    lst_append(lu,p);
    lp_add(ss->clause_best,lu,poss_r,bits,pos,theta,t,0.0);
  } else {
    if (DB(dbinvent)) printf("Don't add it\n");
    fl_free(t->lit_free_list,pos);
    lst_free(lu);
  }
  lp_tree_remove(raw_pred,3);
  lst_free(vt_index);
  lst_free(type_index);

}

void
invent_predicate(search *ss,tab *t, info *i)
{
  /* consider all combination of predicates with up to three terms */

  int a, b;
  lst v_index;
  int num_vars;

  if (invent) {
  if (DB(dbstage)) printf("==Inventing Predicates\n");
  
 
  num_vars = i->num_var_tab;
  v_index = lst_new(3);
  if (DB(dbinvent)) printf("===single var preds inv-num_vars %d\n",num_vars);
 
  for (a=0 ; a < num_vars ; a++) {
    if (DB(dbinvent)) printf("inv-looking at var %d\n",a);
    lst_clear(v_index); lst_append(v_index,a); 
    asses_pred(ss,t,i,v_index);
  }

  if (DB(dbinvent)) printf("===dual var preds inv-num_vars %d\n",num_vars);
 
  for (a=0 ; a < num_vars ; a++) {
    for (b=a+1; b < num_vars; b++) {
      if (DB(dbinvent)) printf("inv-looking at var %d %d\n",a,b);
      lst_clear(v_index); lst_append(v_index,a); lst_append(v_index,b); 
      asses_pred(ss,t,i,v_index);
    }
  }

  /*  if (DB(dbinvent)) printf("===quod var preds inv-num_vars %d\n",num_vars);
 
  for (a=0 ; a < num_vars ; a++) {
    for (b=a+1; b < num_vars; b++) {
      for (c=b+1; c < num_vars; c++) {
	if (DB(dbinvent)) printf("inv-looking at var %d %d\n",a,b,c);
	lst_clear(v_index); lst_append(v_index,a); lst_append(v_index,b); 
	lst_append(v_index,c); 
	asses_pred(ss,t,i,v_index);
      }
    }
  }*/

  lst_free(v_index);



   
  if (DB(dbsearchb)) {      
    printf("\n--- Clause Best List (after predicate invention\n");
    lp_show(ss->clause_best,t,i);
 
  }
  }
}






