/* main.c */

#include <stdlib.h>
#include <string.h>
#include "main.h"


unsigned long debug;
int lst_msize;
int step_size;
int litbase_size;
int c_poss_size;  
int c_best_size;
int c_set_poss_size;
int c_set_final_size;
int var_size;
int mem_use = 0;
int lst_use = 0;
float noise;
int model;
int recursive = 1;
int boosting = 0;
int boostingh = 0;
int boosthval = 0;
int bagging = 0;
int invent = 0;
int noconstant = 0;
float mu_clauses;
float mu_lit;
int num_trials = 10;
float apriorfactor = 1.0;

void
lex_test()
{
	token t;
	while((t = lexan()) != DONE) {
		show_token (t);
	}
	show_token(t);
}

void
ilp(char f[])
{
  info *i;
  tab *t;
  int b, j;
  search *ss;
  boost *bc;
  boost *bagc;
  int pc, np, nc, nn;
  float error;

  mir_check();
  
  lst_msize = 50*litbase_size+
    40*var_size+
    1000*(litbase_size + c_poss_size + 
	 c_best_size + c_set_poss_size + c_set_final_size)+
    1000000;
  if (DB(dbstage)) printf("=lst_msize %d\n",lst_msize);
  lst_create(lst_msize,lst_msize);
  init_rlist();
  init_symtab();
  init(f);
  ps_init();
  if (DB(dbstage)) printf("=Parsing input file.\n");
  parse();
  if (DB(dbstage)) printf("=Learn Type Infomation.\n");
  learn_type();
  fix_pos();
  if (DB(dbstage)) printf("=Generate theta estimation set.\n");
  gen_theta();
  if (DB(dbtrans)) {
    printf("THETA\n");show_pred_set(ps_theta);
  }
  
  
  if (DB(dbtrans)) {
    printf("E+\n");show_pred_set(ps_epos);
    printf("E-\n");show_pred_set(ps_eneg);
    printf("BG\n");show_pred_set(ps_bg);
    printf("T+\n");show_pred_set(ps_tpos);
    printf("T-\n");show_pred_set(ps_tneg);
  }
  if (DB(dbstage)) printf("=Table allocation.\n");
  t = tab_create(ps_count(ps_epos),
		 ps_count(ps_eneg),
		 ps_count(ps_theta),
		 litbase_size + (c_poss_size* BC_DEPTH) + 
		 c_best_size + c_set_poss_size + c_set_final_size,
		 var_size);
  i = info_create(var_size,t);
  info_init(i,4.0);
  ss = init_search(c_poss_size, c_best_size, c_set_poss_size, c_set_final_size);

  if (DB(dbstage)) printf("=Transforming predicates.\n");
  trans(i);
  if (DB(dbtrans)) show_gen_lit(i);
  
  
  if (DB(dbstage)) printf("=Starting table construction.\n");
  b=FALSE;
  j = 0;
  while (!b && 
	 (i->num_lit_tab < litbase_size) && 
	 ((j < step_size) || (step_size == -1))) {
    j++;
    b = info_step(i);
  }
  if (DB(dbtrans)) {
    printf ("step %d res %d lits %d litbase %d\n",
	    j,
	    b,
	    i->num_lit_tab,
	    litbase_size );
  }
  
  if (DB(dbstab)) {
    printf ("Transformed table.\n");
    info_tab_show(t,i);
  }
  if (!boosting && !bagging && !boostingh) { /* no boosting - the default */
    if (DB(dbstage)) printf("=Starting search\n");
    do_search(i,ss);
    
    
    if (DB(dbstage)) printf("=Induced Hypothesis\n");
    if (!DB(dbstage) && DB(dbhyp)) {
      show_lp_nostats (stdout, DIX((ss->clause_set_final),0,lp),i);
    } else {
      if (have_tests) {
	if (DB(dbhyps)) show_lp_lst_test (stdout,ss->clause_set_final,i);
	if (DB(dbhyp)) show_lp_test (stdout, DIX((ss->clause_set_final),0,lp),i);
	if (DB(dbnoisetab)) show_tab_lp_lst(stdout,ss->clause_set_final,i);
      } else {
	if (DB(dbhyps)) show_lp_lst(stdout,ss->clause_set_final,i);
	if (DB(dbhyp)) show_lp_nostats (stdout, DIX((ss->clause_set_final),0,lp),i);
	if (DB(dbnoisetab)) show_tab_lp_lst(stdout,ss->clause_set_final,i);
      }
    }
  } else if (boosting) {  /* doing boosting */
    bc = boost_newc(num_trials);
    boost_do(i,ss,bc,num_trials);
    if (DB(dbhyps) || DB(dbhyp)) {
      boost_test(ps_epos,bc,&pc,&np,i);
      boost_test(ps_eneg,bc,&nc,&nn,i);
      error = error_cal(pc,np,nc,nn);
      printf("=== Data %d/%d   %d/%d  => %6.4f\n", pc, np, nc, nn, error);  
    }
    if (have_tests && (DB(dbhyps) || DB(dbhyp))) {
      boost_test(ps_tpos,bc,&pc,&np,i);
      boost_test(ps_tneg,bc,&nc,&nn,i);
      error = error_cal(pc,np,nc,nn);
      printf("=== Test %d/%d   %d/%d  => %6.4f\n", pc, np, nc, nn, error);
    } 
    if (DB(dbhyps) || DB(dbhyp)) boost_show(bc,i);
    
  } else if (boostingh) {  /* doing boostingh */
    bc = boost_newc(num_trials);
    boosth_do(i,ss,bc,num_trials,boosthval);
    if (DB(dbhyps) || DB(dbhyp)) {
      boost_test(ps_epos,bc,&pc,&np,i);
      boost_test(ps_eneg,bc,&nc,&nn,i);
      error = error_cal(pc,np,nc,nn);
      printf("=== Data %d/%d   %d/%d  => %6.4f\n", pc, np, nc, nn, error);  
    }
    if (have_tests && (DB(dbhyps) || DB(dbhyp))) {
      boost_test(ps_tpos,bc,&pc,&np,i);
      boost_test(ps_tneg,bc,&nc,&nn,i);
      error = error_cal(pc,np,nc,nn);
      printf("=== Test %d/%d   %d/%d  => %6.4f\n", pc, np, nc, nn, error);
    } 
    if (DB(dbhyps) || DB(dbhyp)) boost_show(bc,i);
    
  } else if (bagging) {
    bagc = boost_newc(num_trials);
    bag_do(i,ss,bagc,num_trials);
    if (DB(dbhyps) || DB(dbhyp)) {
      boost_test(ps_epos,bagc,&pc,&np,i);
      boost_test(ps_eneg,bagc,&nc,&nn,i);
      error = error_cal(pc,np,nc,nn);
      printf("=== Data %d/%d   %d/%d  => %6.4f\n", pc, np, nc, nn, error);  
    }
    if (have_tests && (DB(dbhyps) || DB(dbhyp))) {
      boost_test(ps_tpos,bagc,&pc,&np,i);
      boost_test(ps_tneg,bagc,&nc,&nn,i);
      error = error_cal(pc,np,nc,nn);
      printf("=== Test %d/%d   %d/%d  => %6.4f\n", pc, np, nc, nn, error);
    } 
    if (DB(dbhyps) || DB(dbhyp)) boost_show(bagc,i);
  }
}

void
print_help_info () {
  printf("Usage : lime [options] file\n");
  printf("options : \n");
  printf("-d hyp|nohyps|nostage|trans|tabin|tabex|search|searchb|\n");
  printf("   stab|lp|vtab|noisetab|perform|invent|boost|bag\n");
  printf("-f norec|boost|boosth num|bag|invent|noconstant\n");
  printf("-l list storeage size\n");
  printf("-t max number of boost/bag trials\n");
  printf("-s max step size\n");
  printf("-b max number of simple lits\n");
  printf("-p breath of A* search for clauses\n");
  printf("-o number of candidate clauses \n");
  printf("-i breath of A* search for logic programs\n");
  printf("-z number of hypothesis produced\n");
  printf("-v max number of variables\n");
  printf("-n %% of noise\n");
  printf("-a a prior factor\n");
  printf("-h this help info\n");
}


void
scanargs(int argc, char *argv[], int *fpos)
{
  long j;
 // unsigned long t;
  int tt;
  char str[30];

  for (j=1;j<argc;j++) {
    if (argv[j][0] == '-') {
      if (argv[j][1] == 'h') {
	print_help_info();
	exit(1); 
      } else if (argv[j][1] == 'd') {
	sscanf(argv[j+1],"%s",str);
	if (strcmp(str,"hyp") == 0) {debug += 1;}
	else if (strcmp(str,"nohyps") == 0) {debug -= 2;}
	else if (strcmp(str,"nostage") == 0) {debug -= 4;}
	else if (strcmp(str,"trans") == 0) {debug += 8;}
	else if (strcmp(str,"tabin") == 0) {debug += 16;}
	else if (strcmp(str,"tabex") == 0) {debug += 32;}
	else if (strcmp(str,"search") == 0) {debug += 64;}
	else if (strcmp(str,"searchb") == 0) {debug += 128;}
	else if (strcmp(str,"stab") == 0) {debug += 256;}
	else if (strcmp(str,"lp") == 0) {debug += 512;}
	else if (strcmp(str,"vtab") == 0) {debug += 1024;}
	else if (strcmp(str,"noisetab") == 0) {debug += 2048;}
	else if (strcmp(str,"perform") == 0) {debug += 4096;}
	else if (strcmp(str,"invent") == 0) {debug += 8192;}
	else if (strcmp(str,"boost") == 0) {debug += dbboost;}
	else if (strcmp(str,"bag") == 0) {debug += dbboost;}
	j++;
      } else if (argv[j][1] == 'f') {
	sscanf(argv[j+1],"%s",str);
	if (strcmp(str,"norec") == 0) {recursive = 0;}
	else if (strcmp(str,"boost") == 0) {boosting = 1; c_set_final_size = 1;}
	else if (strcmp(str,"bag") == 0) {bagging = 1; c_set_final_size = 1;}
	else if (strcmp(str,"invent") == 0) {invent = 1;}
	else if (strcmp(str,"noconstant") == 0) {noconstant = 1;}
	else if (strcmp(str,"boosth") == 0) {
	  boostingh = 1;
	  c_set_final_size = 1;
	  sscanf(argv[j+1],"%d",&tt);
	  boosthval = tt;
	  j++;
	}
	j++;
      } else if (argv[j][1] == 'a') {
        sscanf(argv[j+1],"%f",&apriorfactor);
        j++;
      } else if (argv[j][1] == 'n') {
        sscanf(argv[j+1],"%f",&noise);
        if (noise < 0.0 || noise >= 1.0) error("Noise should be between in [0.0,1.0)");
        j++;
      } else {
	if (j+1 >= argc) error("Incorrect format : see lime -h"); 
	sscanf(argv[j+1],"%d",&tt);
//	t = (unsigned long) tt;
	switch (argv[j][1]) {
	case 'l' :
	  lst_msize = tt;
	  break;
	case 's' :
	  step_size = tt;
	  break;
	case 'b' :
	  litbase_size = tt;
	  break;
	case 'p' :
	  c_poss_size = tt;
	  break;
	case 'o' :
	  c_best_size = tt;
	  break;
	case 'i' :
	  c_set_poss_size = tt;
	  break;
	case 'z' :
	  c_set_final_size = tt;
	  break;
	case 'v' :
	  var_size = tt;
	  break;
	case 'm' :
	  model = tt;
	  break;
	case 't' :
	  num_trials  = tt;
	  break;
	}
	j++;
      }
    } else {
      *fpos = j;
    }
  }
}


float  CPUTime()
{
    struct rusage usage;

    getrusage(RUSAGE_SELF, &usage);

    return (usage.ru_utime.tv_sec + usage.ru_stime.tv_sec) +
           (usage.ru_utime.tv_usec + usage.ru_stime.tv_usec) / 1E6;
}


int
main (int argc, char *argv[])
{
  int fpos;

  debug = 6;
  step_size = -1;
  litbase_size = 500;
  c_poss_size = 100;  
  c_best_size = 200;
  c_set_poss_size = 50;
  c_set_final_size = 5;
  mu_clauses = 300.0;
  mu_lit = 50.0;
  var_size = 500;
  noise = 0.1;
  model = MOD_FIXED_NOISE;


  if (argc < 2) {
    print_help_info();
    error("file name required");
  }
  
  scanargs(argc,argv,&fpos);
  
  if (DB(dbstage)) {
    printf ("=File : %s\n",argv[fpos]);
    printf ("=Debug level - ");
    if (DB(dbhyp)) printf("hyp ");
    if (DB(dbhyps)) printf("hyps ");
    if (DB(dbstage)) printf("stage ");
    if (DB(dbtrans)) printf("trans ");
    if (DB(dbtabin)) printf("tabin ");
    if (DB(dbtabex)) printf("tabex ");
    if (DB(dbsearch)) printf("search ");
    if (DB(dbsearchb)) printf("searchb ");
    if (DB(dbstab)) printf("stab ");
    if (DB(dblp)) printf("lp ");
    if (DB(dbvtab)) printf("vtab ");
    if (DB(dbnoisetab)) printf("noisetab ");
    if (DB(dbperform)) printf("perform ");
    if (DB(dbinvent)) printf("invent ");
    if (DB(dbboost) && boosting) printf("boost ");
    if (DB(dbboost) && bagging) printf("bag ");
    printf ("\n");
    printf ("=Noise(-n) :%1.2f  litbase(-b) :%d  c_poss(-p) :%d  c_best(-o) :%d\n",
                noise,litbase_size,c_poss_size,c_best_size);
    printf ("=c_set(-i) :%d  c_set_final(-z):%d var_tab:%d apriorfactor(-a) : %1.3f\n",
                c_set_poss_size,c_set_final_size,var_size,apriorfactor);
    if (!recursive) printf("=No recursive clauses.\n");
    if (boosting) printf("=Boosting used. max trials = %d\n",num_trials);
    if (boostingh) printf("=Hyb %d Boosting used. max trials = %d\n",boosthval,num_trials);
    if (bagging) printf("=Bagging used. max trials = %d\n",num_trials);
    if (invent) printf("=Predicate invention used.\n");
    if (noconstant) printf("=Constants not used as output .\n");
  }
  if (DB(dbnoisetab)) {
    printf ("%s\t",argv[fpos]);
    printf ("%1.2f\t",noise);
  }
  ilp(argv[fpos]);

  if (DB(dbstage)) {
    printf("Seconds %10.2f\n",CPUTime());
    printf ("Exit\n");
  }


  return 0;

}







