/* main.h */

/*#define SUN*/
#define LINUX

#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <math.h>

#ifdef LINUX
#include <values.h>
#endif

#define REAL_TYPE 0
#define INT_TYPE 1
#define SS_TYPE 2


#define MAXSIMPLECLAUSE 40

#define TRUE 1
#define FALSE 0
#define NEXT 1
#define LAST -1
#define MAX_VAR 25
#define MAXARITY 25
#define PRED_START 100

#define THETASIZE 500

#define FC_DEPTH 2
#define BC_DEPTH 15

#define BIGNEG -9999999.0


#define log2(A) (log(A)/M_LN2)

extern int model;
extern int recursive;
extern int noconstant;
extern int invent;
extern int boosting;

#define MOD_SIM          0 
#define MOD_SIM_NOISE    1
#define MOD_FIXED        2
#define MOD_FIXED_NOISE  3


extern unsigned long debug;      
#define DB(A)      (debug & A)
#define dbhyp      1 
#define dbhyps     2 
#define dbstage    4 
#define dbtrans    8 
#define dbtabin    16 
#define dbtabex    32 
#define dbsearch   64 
#define dbsearchb  128 
#define dbstab     256 
#define dblp       512
#define dbvtab     1024 
#define dbnoisetab 2048
#define dbperform  4096
#define dbinvent   8192
#define dbboost    16384

extern int mem_use;
extern int lst_use;
extern int lst_msize;
extern float noise;
extern int numentry;
extern float mu_clauses;
extern float mu_lit;
extern int symtmap;
extern float apriorfactor;


#include "bits.h"
#include "util.h"
#include "map.h"
#include "table.h"
#include "info.h"
#include "search.h"
#include "input.h"
#include "trans.h"
#include "perform.h"
#include "boost.h"

/* main */

void scanargs(int , char*[], int*);

/* bits - bit vecter set */ 


bvs *bvs_create(int,int); /* bits, vectors */
int bvs_count_bits (bvs*,int);
void bvs_intersection (bvs*,int,int,int); /* in1, in2, out*/
void bvs_and (bvs*,int,int,int); /* in1, in2, out*/
void bvs_or (bvs*,int,int,int); /* in1, in2, out*/
void bvs_copy (bvs*,int,int); /* in1, out*/
void bvs_clear (bvs*,int,int);   /* vector, value */
void bvs_set (bvs*,int,int,int); /* vector, bit, value */
int bvs_get (bvs*,int,int); /* vector, bit */
void bvs_show (bvs*);
void bvs_show_index (lst index, bvs *bvsp);
float bvs_dot_f (bvs*,int,float[]);
void bvs_Union(bvs*,int,int[],int); /* num, list, out */
int bvs_leq(bvs*,int,int);
int bvs_eq(bvs*,int,int);
int bvs_hash_fun(bvs*,int); /* pos */
void bvs_test();

/* util */

float my_random_float (float b);
int my_random(int); /* n */
void *memory (int,char[]);
void mem_free(void*);
void error(char[]);
void arr_cp (int,int[],int,int[],int); /* num,in[],in offset, out[],
										  out offset */
int arr_cmp (int, int[], int, int[], int); /* num, a[], a offset, b[],
											  b offset */
int arr_cmp_lt (int, int[], int, int[], int); /* num, a[], a offset, b[],
												 b offset */
fl *fl_create(int); /* size */
void fl_clear(fl *f);
void fl_dispose (fl *f);
int fl_full(fl*);
int fl_get(fl*);
void fl_free(fl*,int);
void lst_create(int, int); /* index_size, data_size */
lst lst_new(int); /* size */
void lst_grow(lst); 
void lst_insert(lst,int,int); /* idx, d, i */
void lst_fill(lst,int,int); /* idx, num, d */
int lst_append(lst, int); /* idx, d */
int lst_look(lst,int); /* idx, i */
int lst_size(lst); /* idx */
void lst_free(lst); /* idx */
int lst_pop(lst); /* idx */
int lst_find(lst,int); /* idx d */
int  lstlst_find(lst x,lst xs);
int  lstlst_member(lst x,lst xs);
int lst_member(int,lst); /* d idx */
int lst_cmp(lst,lst); /* l1 l2 */
void lst_shift(lst,int,int); /* idx d i */
void lst_show(lst); /* idx */
void bv_to_lst(bvs*,lst,int); /* s res lit */
void lst_clear(lst); /* idx */
void lst_copy (lst, lst ); /* s d */
int lst_sum(lst d);
int lst_head(lst); /* idx */
int lst_subset(lst ss, lst s);
void lst_merge(lst,lst,lst); /* s1 s2 d */
void lst_2_arr(lst l,int a[]);
lst lst_replicate(lst x);
void arr_2_lst(int a[], int num, lst l);
void init_rlist ();
int look_rlist (int); /* index */
int lst_max(lst l, int a);
int max(int a, int b);
int lst_del(lst idx,int i);

float mir_float(int i);
int mir_int(float f);
void mir_check();

/* map */

map *map_create(int,int,int); /* max_num_el, in, out */
void map_add(map*,int[],int[],int); /*map, intput, output idx*/
int map_hash(map*, int[]); /* m in */
int map_lookup_lst(map *m, lst in, lst out, lst *idx);
int map_lookup(map*,int[],int[],lst*); /*map, intput, output idx*/
void map_test();
void map_free(map*); /* map */
int map_cmp(map*,map*); /* m1, m2 */
int map_add_sort(map*,int[],int[],int); /* map, input, output, idx */
					/*  ret TRUE ok fun */
void map_show(map*); /* m */


/* table */

tab *tab_create(int,int,int,int,int);/* num_e_p, num_e_n, num_theta,
					    num_max_total_lit, num_max_vars */

int  tab_bits_hash_fun(tab*,int ); /* t pos */
void tab_bits_hash_add(tab *t, int pos, int h); /* t pos */
int tab_bits_hash_find(tab *t, int pos, int h); /* t pos */
int tab_bits_hash_notfindadd(tab *t, int pos);
int  tab_vec_hash_fun(tab*,int ); /* t pos */
void tab_vec_hash_add(tab*, int ); /* t pos */
int tab_vec_hash_find(tab*, int ); /* t pos */
void tab_hash_bits_clear(tab *t);
void tab_add_examples(tab*,int[],int*);
		/* t, num_terms, examples, term_place, lit_place */
void tab_new_lit(tab*,int,int[],int,int[],map*,int*,int);
	  /* t,num_in_vars, in_vars, num_out_vars, out_vars,
		 function, lit_pos is_rec*/
void tab_free_lit(tab*,int); /* t pos */
void tab_free_var(tab*,int); /* t pos */
void tab_show(tab*);
void info_tab_show(tab *t, info *i);
int tab_replicate(tab *t, int lit);
float tab_bits(tab*, int); /* t lit */
void cal_cover(tab*,int); /* t, lit */
void no_cover(tab*,int); 
void tab_cover(tab*,int,float*,float*,float*); /* t, lit pc nc tc*/
void tab_copy(tab *t, int lit, int litp);
void tab_clear(tab *t, int lit);


/* info */

info *info_create (int, tab *);
		 /* max_var_tab, *the_tab */
void info_init (info *, float );
		/* info *i, int num_terms, int examples[], int type[], float bits */
void info_add_genlit (info *, int, int, glit (*)[]); /* i, max, num, gen_lit*/
int info_step (info *); /* i */
void info_test ();
void info_getlit(info*, lst, int); /* i res lit */
int prev_lit_better(info*,int); /*  i l_pos */
int prev_var_same(info*,int); /*  i var no */
float info_bits(info *,lst); /* i, lits */ 

/* search */

lp_lst *lp_create(int, int ); /* size ,lev */
void lp_reset(lp_lst *lp, tab *t);
void lp_show_j (lp_lst *lp, tab *t, info *i, int j); /* lp t i j */
void lp_show (lp_lst *lp, tab *t, info *i); /* lp t i */
void lp_tree_remove (lst, int); /* idx lev */
void lp_copy(lst,lst); /* s d */
lst lp_bestlp(lp_lst *lp);
int lp_bestlit(lp_lst *lp);
float lp_best(lp_lst*); /* lp */
float lp_worse(lp_lst*); /* lp */
int lp_size(lp_lst*);
float lp_bits(lp_lst *lp, int pos);
float lp_bay(lp_lst *lp, int pos);
void lp_head(lp_lst*,lst*,float*,int*,float*); /* lp p bits lit last_bay*/
int lp_add(lp_lst*, lst, float, float, int, float, tab*, float);
		/* lp p v1 v2 lit  theta t last_bay */
float bay(tab *t, float bits, float pc, float nc, float theta); 
float best_bay(tab *t, float bits, float pc, float nc, float theta);
float best_and_bay(tab *t, float bits, float pc, float nc, float theta);
float prob_lp (lst lp, info *i);
float prob_clause(lst c, info *i);
float log_geo (int n, float mu);
void do_search(info*,search*); /* ss i */
search *init_search(int gain_list_size, 
		    int max_clause_best, 
		    int max_clause_set_possible, 
		    int max_clause_set_final);
void add_simple_clause_to_pos (search *ss, tab *t, info *i);
void form_clauses(search *ss, tab *t, info *i);
void add_clause_to_pos (search *ss,tab *t,info *i);
void form_programs(search *ss,tab *t,info *i);
float cal_theta(tab *t, float tc);
void probe_clause(tab *t, info *i, search *ss, int c, lst cp, int lev, lst gain_order_ranking, int prevj);
void lp_pop(lp_lst *lp,tab *t);
lst get_gain_order_ranking(tab *t, info *i);


/* input */

void init_symtab();
void show_token(token); /* t */
void init(char[]); /* file name */
token lexan();
void l_error(char[]); /* message */
int lookup(char[]); /* string */
int insert(char[]); /* string */
void parse();
void p_epos();
void p_eneg();
void p_tpos();
void p_tneg();
void p_bg();
void p_perd_list(pred_set*); /* ps */
void p_perd(pred_set*); /* ps */
void p_term(token*,int*); /* t , v */
int match(token);
void show_lp_nostats(FILE *f, lst lp, info *i);
void show_lp_lst (FILE*, lp_lst*, info*); /* f lpl i */
void show_lp_lst_test (FILE*, lp_lst*, info*); /* f lpl i */
void show_tab_lp_lst (FILE*, lp_lst*, info*); /* f lpl i */
void show_lp (FILE*, lst, info*); /* f lp i */
void show_lp_test (FILE*, lst, info*); /* f lp i */
void show_lp_nnl (FILE*, lst, info*); /* f lp i */
void show_c (FILE*, lst, info*); /* f c i */
void show_l (FILE*, int, info*); /* f l i */
int all_rec_lits_lp(lst lp, info *i);
int no_rec_lits_lp(lst lp, info *i);
int no_rec_lits_c(lst c, info *i);
int no_rec_lit(int l, info *i);
/* trans */

void ps_init();
void create_tt (int num_types);
void add_tt (int ot, int val);
int get_tt (int ot);
void free_tt ();
int type_spreed(lst tl, int re_num);
void add_each_type(lst tl, lst pds);
void reassign_type(lst tl);
void learn_type();
void fix_pos();
void add_same(lst el_list,int kt,lst preds,lst type);
void gen_theta();
pred_set *ps_create();
int ps_count(pred_set*); /* ps */
void ps_add(pred_set*, int, int, lst, lst); /* ps, name, arity, tp, d */
void trans(info*);
void show_pred_set(pred_set*); /* ps */
void show_gen_lit(info*); /* i */
void show_an_example (lst data, lst type); 

/* perform */
int calculate_cover(info *i, lst hyp, pred_set *ps);
int cover_ex2(info *i, lst hyp, lst data);
void cover(info *i, lst hyp, int *steps, int lpos);
int cover_ex(info *i, lst hyp, int *step, int tpos);

int resolve(info *i, 
	    int desent, 
	    int hint,
	    int raw_pred, 
	    lst data,  
	    lst hyp, 
	    int *st, 
	    int *hypnum);
int  uni_res(info *i, int desent, lst data, lst hyp, int *st, lst c);
int res_maxvar(info *i,lst cla, int lit);
int hint_okay(info *i, lst data, int lit);
int find_pred(info *i, lst data, int raw_pred);
int bgresolve(info *i, int pred, lst data);

/* boost */
float error_cal (float pc, float np, float nc, float nn);
void boost_test(pred_set *ps, boost *bc,int *c,int *n,info *i);
boost *boost_newc(int num_trials);
void boost_do (info *i,search *ss, boost *bc, int num_trials);
void boost_show (boost *bc, info *i);
void boosth_do (info *i,search *ss, boost *bc, int num_trials, int boosthval);
void  bag_do (info *i,search *ss, boost *bc, int num_trials);

/* invent */
void invent_predicate(search *ss,tab *t, info *i);


#define DIX(P,I,V) (((*P->data)[lst_look(P->index,I)]).V)


#ifdef SUN
#define BITSPERBYTE     8

#define BITS(type)      (BITSPERBYTE * (int)sizeof(type))
#define HIBITS  ((short)(1 << BITS(short) - 1))
#define HIBITI  (1 << (BITS(int) - 1))
#define HIBITL  (1L << BITS(long) - 1)


#define MAXSHORT        ((short)~HIBITS)
#define MAXINT  (~HIBITI)
#define MAXLONG (~HIBITL)
#endif
  




