/* map.c */

#include "main.h"

map
*map_create(int max, int in, int out)
{
  map *m;
  register int mem_req;
  register int s, j, mmax;
  
  mmax = 3 * max + 3;
  mem_req = sizeof(int) * mmax * (s = (in + out));
  m = (map *) memory (sizeof(map),"create map struct");
  m->data = (int (*)[]) memory(mem_req, "creating data for map");
  m->index = (int (*)[]) memory(sizeof(int) * mmax,"map index");
  m->max_num_el = mmax;
  m->num_el = 0;
  m->num_in = in;
  if (in > 10) error ("hash too big");
  m->num_out = out;
  m->num_size = s;
  for (j=0;j<m->max_num_el;j++) {
    (*m->index)[j] = lst_new(3);
  }
  return(m);
}

void
map_add(map *m, int in[], int out[], int idx)
{
  register int j, hashp;
  register int pos;

  if (m->num_el >= m->max_num_el-2) error("over done map");
  hashp = map_hash(m,in);
  for (j=hashp;lst_size((*m->index)[j]) != 0;j = (j+1) % m->max_num_el) ;
  pos = j * m->num_size;
  arr_cp(m->num_in,in,0,(*m->data),pos);
  arr_cp(m->num_out,out,0,(*m->data),pos + m->num_in);
  lst_append((*m->index)[j],idx);
  m->num_el++;
}


int
map_hash(map *m, int in[])
{
  int t, j;

  t = 0;
  for (j=0;j<m->num_in;j++) {
    t += look_rlist(j)*in[j] % 6000;
  }
  if (t < 0) t = -t;
  return (t % m->max_num_el);
}


int
map_lookup_lst(map *m, lst in, lst out, lst *idx)
{
  int tm1[MAX_VAR];
  int tm2[MAX_VAR];
  int found;

  lst_2_arr(in,tm1);
  found = map_lookup(m,tm1,tm2,idx);
  if (found) {
    arr_2_lst(tm2,m->num_out,out);
  }
  return found;
}

int
map_lookup(map *m, int in[], int out[], lst *idx)
{
  int j, found;
  int hashp;

  hashp = map_hash(m,in);
  found = FALSE;
  j = hashp;
  while ((lst_size((*m->index)[j]) != 0) &&
         !(found = arr_cmp(m->num_in,in,0,(*m->data),j * (m->num_size)))) {
    j = (j+1) % m->max_num_el;
  }

  if (found) {
    arr_cp(m->num_out,(*m->data),j*m->num_size + m->num_in,out,0);
    *idx = (*m->index)[j];
    return (TRUE);  
  } else {
    return (FALSE);
  }
}


void
map_free(map *m)
{
	mem_free(m->data);
	mem_free(m->index);
	mem_free(m);
}

int
map_add_sort(map *m, int in[], int out[], int idx)
{
  int j, hashp;
  int pos, found;


  if (m->num_el >= m->max_num_el-2) error("over done map");
  hashp = map_hash(m,in);
  found = FALSE;
  j = hashp;
  while ((lst_size((*m->index)[j]) != 0) &&
         !(found = arr_cmp(m->num_in,in,0,(*m->data),j * (m->num_size)))) {
    j = (j+1) % m->max_num_el;
  }

  if (found) {
    if (arr_cmp(m->num_out,out,0,(*m->data),j*m->num_size + m->num_in)) {
      lst_append((*m->index)[j],idx);
      return TRUE;
    } else {
      return FALSE;
    }    
  } else {
    pos = j * m->num_size;
    arr_cp(m->num_in,in,0,(*m->data),pos);
    arr_cp(m->num_out,out,0,(*m->data),pos + m->num_in);
    lst_append((*m->index)[j],idx); 
    m->num_el++;
    return TRUE;
  }
}

int
map_cmp(map* m1,map *m2)
{
  int i, j;
  int hashp, found;

  if ((m1->num_el != m2->num_el) ||
      (m1->num_in != m2->num_in) ||
      (m1->num_out != m2->num_out) ) return FALSE;
  
  for (i=0;i<m1->max_num_el;i++) {
    if ((*m1->index)[i] != -1) {
      hashp = map_hash(m2,&((*m1->data)[i*m1->num_size]));

      found = FALSE;
      j = hashp;
      while ((lst_size((*m2->index)[j]) != 0) &&
	     !(found = arr_cmp(m2->num_in,&((*m1->data)[i*m1->num_size]),0,
			       (*m2->data),j*(m2->num_size)))) {
	j = (j+1) % m2->max_num_el;
      }
      
     
      if (found) {
	if (!arr_cmp(m1->num_out,(*m1->data),i*(m1->num_size)+m1->num_in,
			         (*m2->data),j*(m2->num_size)+m2->num_in)) 
	  return FALSE;
      } else {
	return FALSE;
      }
    }
  }
  return TRUE;
}


void
map_show(map* m)
{
  int i, j;

  for (i=0;i<m->max_num_el;i++) {
    if (lst_size((*m->index)[i]) != 0) {
      for(j=0;j<m->num_in;j++) printf("%d ",(*m->data)[j+i* m->num_size]);
      printf("->");
      for(j=0;j<m->num_out;j++)
	printf("%d ",(*m->data)[j+i* m->num_size + m->num_in]);
      printf("\n");
    }
  }
}

/*
void
map_test ()
{
map *mp;
int a[] = {1,2,3};
int ao[] = {5,6};
int b[] = {3,2,3};
int bo[] = {51,62};
int c[] = {5,11,2};
int co[] = {15,16};
int d[] = {22,15,16};
int r;
int ro[2];
int i;

	mp = map_create(20,3,2);
	map_add (mp,a,ao,1);
	map_add (mp,b,bo,2);
	map_add (mp,c,co,3);
	r = map_lookup(mp,a,ro,&i);
	printf("result %d, output %d, %d \n",r,ro[0],ro[1]);
	if ((r == 1) && (ro[0] == ao[0]) && (ro[1] == ao[1])) printf("okay\n");
	r = map_lookup(mp,b,ro,&i);
	printf("result %d, output %d, %d \n",r,ro[0],ro[1]);
	if ((r == 1) && (ro[0] == bo[0]) && (ro[1] == bo[1])) printf("okay\n");
	r = map_lookup(mp,c,ro,&i);
	printf("result %d, output %d, %d \n",r,ro[0],ro[1]);
	if ((r == 1) && (ro[0] == co[0]) && (ro[1] == co[1])) printf("okay\n");
	r = map_lookup(mp,d,ro,&i);
	printf("result %d, output %d, %d \n",r,ro[0],ro[1]);
	if (r == 0) printf("okay\n");

}
*/



