/* perform.c */

#include "main.h"

#define DESENTLEVEL 14 /* was set at 13 */
#define MAXRESSTEPS 50

void
cover(info *i, lst hyp, int *steps, int lpos)
{
  int total_step,step,tpos, k;
  float num_cover;
  tab *t;

  
  if (DB(dbperform) || DB(dbsearch)) {
    printf ("==== Determine performance and coverage\n");
    show_lp_nostats (stdout, hyp, i);
  }

  t = i->the_tab;
  total_step = 0;

  num_cover = 0.0;                         /* calculate positive cover */
  for (k = 0; k < t->num_e_p; k++) {
    tpos = k;
    if (cover_ex(i,hyp,&step,tpos)) {
      num_cover += (*t->weight_p)[k];
      bvs_set(t->lit_cover_p,lpos,k,1);
    } else {
      bvs_set(t->lit_cover_p,lpos,k,0);
    }
    total_step += step;
  }
  (*t->cover_p)[lpos] = num_cover;

  num_cover = 0.0;                         /* calculate negative cover */
  for (k = 0; k < t->num_e_n; k++) {
    tpos = t->num_e_p + k;
    if (cover_ex(i,hyp,&step,tpos)) {
      num_cover += (*t->weight_n)[k];
      bvs_set(t->lit_cover_n,lpos,k,1);
    } else {
      bvs_set(t->lit_cover_n,lpos,k,0);
    }
    total_step += step;
  }
  (*t->cover_n)[lpos] = num_cover;

  num_cover = 0.0;                         /* calculate theta cover */
  for (k = 0; k < t->num_theta; k++) {
    tpos = t->num_e_n + t->num_e_p + k;
    if (cover_ex(i,hyp,&step,tpos)) {
      num_cover += (*t->weight_t)[k];
      bvs_set(t->lit_cover_t,lpos,k,1);
    } else {
      bvs_set(t->lit_cover_t,lpos,k,0);
    }
    total_step += step;
  }
  (*t->cover_t)[lpos] = num_cover;

  *steps = total_step;

  
  if (DB(dbperform) ||  DB(dbsearch)) {
    printf ("==== Done p %3.3e n %3.3e t %3.3e\n",(*t->cover_p)[lpos],(*t->cover_n)[lpos],(*t->cover_t)[lpos]);
  }

}

int
calculate_cover(info *i, lst hyp, pred_set *ps)
{
  int k,arity;
  int sum;
  lst data;
  
  if (DB(dbperform) ||  DB(dbsearch)) {
    printf ("==== Determine performance and coverage\n");
    show_lp_nostats (stdout, hyp, i);
  }


  sum = 0;
  if (ps->num != 0) {
    arity = ps->preds[0].arity;
    data = lst_new(arity);
    for (k=0;k<(ps_count(ps));k++) {
      lst_copy(lst_look(ps->preds[0].preds,k),data);
      if (cover_ex2(i,hyp,data)) sum += lst_look(ps->preds[0].number,k);
      /* if (DB(dbperform)) {
	printf("(%1d)",lst_look(ps->preds[0].number,k));
	show_an_example(data,ps->preds[0].type);
	printf("\n");
      }*/
    }
    lst_free(data);
  }

  
  if (DB(dbperform) ||  DB(dbsearch)) {
    printf ("==== Done Sum %d\n",sum);
  }

  return sum;
}

int
cover_ex2(info *i, lst hyp, lst data)
{
  int st, hypnum;
  int result;

  st = 0;


  result = resolve(i, DESENTLEVEL, 0, -1,
		   data,
		   hyp,&st,&hypnum);
  
  /*
  if (DB(dbperform)) {
    if (result)
      printf ("Yes %4d %4d ",st,hypnum);
    else
      printf ("No  %4d %4d ",st,hypnum);
  }*/

  return result;
}

int
cover_ex(info *i, lst hyp, int *step, int tpos)
{
  tab *t;
  pred *p;
  int k, hypnum, st, result, lit_name;
  lst data;
  


  /* get data from table into list */
  t = i->the_tab;

  p = &(ps_epos->preds[0]);


  data = lst_new(p->arity);
  for (k=0;k<p->arity;k++)
    lst_append(data,((*t->var_table)[(*i->var_tab)[k].pos_in_tab * (t->num_et) + tpos]));
	       
  /*  if (DB(dbperform)) {
    printf ("===== Hyp cover ( ");
    show_an_example(data,p->type);
    printf (" ) ? \n");
  }*/



  /* set lit name to 0 */
  lit_name = 0;

  st = 0;

  /* call resolve */
  
  result = resolve(i, DESENTLEVEL, lit_name, -1, 
		   data,
		   hyp,&st,&hypnum);
  
  /* free lists */
  lst_free(data);

  /* set steps */
  *step = st;
  /*
  if (DB(dbperform)) {
    if (result)
      printf ("Yes %d %d \n",st,hypnum);
    else
      printf ("No  %d %d \n",st,hypnum);
    }*/


  return result;

}


void 
indent(int i) {
  int k;
  for(k=0;k<i;k++) printf(" ");
}

int
resolve(info *i, 
	int desent, 
	int hint,
	int raw_pred, 
	lst data,  
	lst hyp, 
	int *st, 
	int *hypnum)
{
  int p, found,k;
  pred *pp;

  (*st)++;
  desent--;



  if (DB(dbperform)) {
    indent(DESENTLEVEL - desent);
    if (raw_pred == -1) {
      pp = &(ps_epos->preds[0]);
      printf("%s",(*symtable)[lst_look(symtmap,pp->name)].lexptr);
      show_an_example(data,pp->type);
      printf("{\n");
    } else {
      pp = &(ps_bg->preds[raw_pred]);
      printf("%s",(*symtable)[lst_look(symtmap,pp->name)].lexptr);
      show_an_example(data,pp->type);
      printf("{");
    }
  }

  if (desent < 0 || (*st > MAXRESSTEPS)) {
    if (DB(dbperform)) {
      printf("}F\n");
    }
    return FALSE;
  }


  if (raw_pred == -1) { /* resolving target predicate */

    for (k=0;k<lst_size(hyp);k++) {
      if (no_rec_lits_c(lst_look(hyp,k),i)) {
	found = uni_res(i,desent,data,hyp,st,lst_look(hyp,k));
	if (found) {
	  *hypnum = k;
	  if (DB(dbperform)) {
	    indent(DESENTLEVEL - desent);
	    printf("}T\n");
	  }
	return TRUE;
	}
      }
    }
    for (k=0;k<lst_size(hyp);k++) {
      if (!no_rec_lits_c(lst_look(hyp,k),i)) {
	found = uni_res(i,desent,data,hyp,st,lst_look(hyp,k));
	if (found) {
	  *hypnum = k;
	  if (DB(dbperform)) {
	    indent(DESENTLEVEL - desent);
	    printf("}T\n");
	  }
	  return TRUE;
	}
      }
    }
    *hypnum = -1;
    if (DB(dbperform)) {
      indent(DESENTLEVEL - desent);
      printf("}F\n");
    }
    return FALSE;    

 
  } else { /* resolving background knowledge predicate */
    if (hint_okay(i,data,hint)) {
      p = hint;
    } else {
      p = find_pred(i,data, raw_pred);
      if (p == -1) {
	*hypnum = -1;
	if (DB(dbperform)) {
	  printf("}F\n");
	}
	return FALSE;
      }
    }
    found = bgresolve(i,p,data);
    *hypnum = -2;
    if (DB(dbperform)) {
      if (found) {
	printf("}T\n");
      } else {
	printf("}F\n");
      }
    }
    return found;
  }
}

int
hint_okay(info *i, lst data, int lit)
{
  lst inrb;
  int k, okay;

  inrb = (*i->gen_lit)[lit].raw_in_bind;
  okay = TRUE;
  for (k=0;k<lst_size(inrb);k++) okay = okay && (lst_look(data,(lst_look(inrb,k))) != -1);
  return okay;
}

int
find_pred(info *i, lst data, int raw_pred)
{
  int k;
  int found;

  found = -1;
  for (k=0;k<i->num_gen_lit && (found == -1);k++) {
    if ((*i->gen_lit)[k].raw_pred == raw_pred) {
      if (hint_okay(i,data,k)) found = k;
    }
  }
  return found;
}


int 
uni_res(info *i, int desent, lst data, lst hyp, int *st, lst c)
{
  int a, max_var, k, resolved, lit, lit_num, raw_pred, invsize, outvsize,j, val, hypnum;
  lst var_sub, inv, outv, inrb, outrb, d;


  /* set up the var sub */
  a = lst_size(data);
  max_var = res_maxvar(i,c,a);
  var_sub = lst_new(max_var+1);
  for (k=0;k<a;k++) lst_append(var_sub,(lst_look(data,k)));
  for (k=a;k<max_var+1;k++) lst_append(var_sub,-1);

  /* loop through each literal in c */
  resolved = TRUE;
  k = 1;
  while (resolved && (k < lst_size(c))) {
    /* set input for literal, make shore it is okay */
    lit = lst_look(c,k);
    lit_num = (*i->lit_tab)[lit].lit_num;
    raw_pred = (*i->gen_lit)[lit_num].raw_pred;
    inv = (*i->lit_tab)[lit].in_vars;
    outv = (*i->lit_tab)[lit].out_vars;
    inrb = (*i->gen_lit)[lit_num].raw_in_bind;
    outrb = (*i->gen_lit)[lit_num].raw_out_bind;
    invsize = lst_size(inv);
    outvsize = lst_size(outv);
    d = lst_new(invsize + outvsize);
    /*printf("%d[",k);*/
    for (j=0;j<invsize;j++) {
      val = lst_look(var_sub,lst_look(inv,j));
      lst_insert(d,val,lst_look(inrb,j));
    }
    for (j=0;j<outvsize;j++) {
      val = -1;
      lst_insert(d,val,lst_look(outrb,j));
    }
    /* resolve the literal */
    if (resolved) {
      resolved = resolve(i, desent, lit_num, raw_pred, d,  hyp, st, &hypnum);
      
      
      /* set the ouput into thet var binding table, make shore its okay */
      if (resolved) {
	for (j=0;j<outvsize;j++) {
	  val = lst_look(d,lst_look(outrb,j));
	  lst_insert(var_sub,val,lst_look(outv,j));
	}
	for (j=0;j<invsize;j++) {
	  val = lst_look(d,lst_look(inrb,j));
	  lst_insert(var_sub,val,lst_look(inv,j));
	}
      }
      k++;
    }
    /*    printf("]");*/
    lst_free(d);
  }

  if (resolved) {   /* fix output data */ /* must fix for head */
    for (k=0;k<a;k++) lst_insert(data,lst_look(var_sub,k),k);
  }

  lst_free(var_sub);
  if (resolved) {
    return TRUE;
  } else {
    return FALSE;
  }
}


int
bgresolve(info *i, int pred, lst data) 
{
  lst data_in, data_out, rib, rob;
  int idx, found, j, d, p, nd;

  rib = (*i->gen_lit)[pred].raw_in_bind;
  rob = (*i->gen_lit)[pred].raw_out_bind;
  data_in = lst_new((*i->gen_lit)[pred].num_in);
  data_out = lst_new((*i->gen_lit)[pred].num_out);

  for (j=0;j<lst_size(rib);j++) {
    lst_append(data_in,lst_look(data,lst_look(rib,j)));
  }

  found = map_lookup_lst((*i->gen_lit)[pred].fun,data_in,data_out,&idx);

  if (found) {
    for (j=0;j<lst_size(rob) && found;j++) {
      p = lst_look(rob,j);
      d = lst_look(data,p);
      nd = lst_look(data_out,j);
      if (d == -1 || d == nd) {
	lst_insert(data,nd,p);
      } else {
	found = FALSE;
      }
    }
  }

  lst_free(data_out);
  lst_free(data_in);
  return found;
}

int
res_maxvar(info *i,lst cla, int a)
{
  int m, k;

  m = a;
  for (k=1;k<lst_size(cla);k++) {
    m = lst_max((*i->lit_tab)[lst_look(cla,k)].in_vars,m);
    m = lst_max((*i->lit_tab)[lst_look(cla,k)].out_vars,m);
  }
  return m;
}


















