/* search.c */

#include "main.h"


#define NUM_LP 5
#define DATAIDX(X) ((*lp->data)[lst_look(lp->index,X)])

#define LOOP_BOUND_C 1000
#define LOOP_BOUND_P 1000


lp_lst
*lp_create(int size, int lev)
{
	lp_lst *lp;

	lp = (lp_lst *) memory(sizeof(lp_lst),"creating search block");
        lp->index = lst_new(size);
	lp->data = (lp_data (*)[]) memory (sizeof(lp_data)*size,"data for lp");
        lp->data_fl = fl_create(size);
	lp->size = size;
	lp->lev = lev;
	return (lp);
}

void
lp_reset(lp_lst *lp, tab *t) 
{
  int num, j, p, lit;
  
  num = lst_size(lp->index);
  for(j=0;j<num;j++) {
    p = DATAIDX(j).lp;
    lit = DATAIDX(j).lit;
    if (p != -1) lp_tree_remove(p,lp->lev);
    if (lit != -1) fl_free(t->lit_free_list,lit);
  } 
  lst_clear(lp->index);
  fl_clear(lp->data_fl);
}

void
lp_show_j (lp_lst *lp, tab *t, info *i, int j)
{
  float pc, nc, tc;
  
  tab_cover(t,(*lp->data)[lst_look(lp->index,j)].lit,&pc,&nc,&tc);
  printf("%4d  %6d  %6d  %6.2f  %6.2f  %6.3f  %6.2f  %6.2f/%6.2f  %6.2f/%6.2f  ",
         j,	 
	 DATAIDX(j).lp,
	 DATAIDX(j).lit,
	 DATAIDX(j).v1,
	 DATAIDX(j).v2,
	 DATAIDX(j).theta,
	 DATAIDX(j).last_bay,
	 pc,
	 t->num_w_p,
	 nc,
	 t->num_w_n);
  if (lp->lev == 2) {
    printf("\n");
    show_lp (stdout, DATAIDX(j).lp, i);
  } else if (lp->lev == 1) {
    show_c (stdout, DATAIDX(j).lp, i);
    printf("\n");
  } else if (lp->lev == 0) {
    show_c(stdout,(*i->lit_tab)[DATAIDX(j).lit].lits_used,i);
    printf("\n");    
  } else {
    printf("\n");    
  }
  
}

void
lp_show (lp_lst *lp, tab *t, info *i)
{
  int j, num;

  printf("size %d lev %d\n",lp->size,lp->lev);
  printf("   j      lp     lit      v1       v2   theta   last    pc/np   nc/nn\n");
  num = lst_size(lp->index);
    for(j=0;j<num;j++) {
      lp_show_j (lp,t,i,j);
    }
 
}

void
lp_tree_remove (lst idx, int lev)
{
	int i;

	if (lev == 2 || lev == 1)
		lst_free(idx);
	if (lev > 2) {
		for (i=0;i<lst_size(idx);i++)
			lp_tree_remove(lst_look(idx,i),lev-1);
		lst_free(idx);
	}
}

void
lp_copy(lst s, lst d)
{
	int j, num;
	lst c, k;

	lst_clear(d);
	num = lst_size(s);
	for (j=0;j<num;j++) {
		c = lst_look(s,j);
		k = lst_new(lst_size(c));
		lst_copy(c,k);
		lst_append(d,k);
	}

}

void
lp_pop(lp_lst *lp, tab *t) 
{
  lst p;
  int lit;

  p = DATAIDX(0).lp;
  lit = DATAIDX(0).lit;
  if (p != -1) lp_tree_remove(p,lp->lev);
  if (lit != -1) fl_free(t->lit_free_list,lit);  
  fl_free(lp->data_fl, lst_look(lp->index,0));
  lst_head(lp->index);   
}

void
lp_head(lp_lst *lp, lst *p, float *bits, int *lit, float *last_bay)
{
	*lit = DATAIDX(0).lit;
        *p = DATAIDX(0).lp;
	*bits = DATAIDX(0).v2;
	*last_bay = DATAIDX(0).last_bay;
        fl_free(lp->data_fl, lst_look(lp->index,0));
	lst_head(lp->index);    
}

void
lp_goodsmall(lp_lst *lp, float bestf, lst *p, float *bits, int *lit, float *last_bay)
{
  int k, best, size;
  float bestv;

  size = lp_size(lp);
  best = 0;
  bestv = DATAIDX(best).v2;
  k = 0;
  while ((k < (size-1)) && (DATAIDX(k).v1 > bestf)) {
    k++;
    if (DATAIDX(k).v2 < bestv) {
	bestv = DATAIDX(k).v2;
	best = k;
    }
  }
  
  /*printf ("k %d size %d best %d bestv %e\n",k,size,best,bestv);*/
  *lit = DATAIDX(best).lit;
  *p = DATAIDX(best).lp;
  *bits = DATAIDX(best).v2;
  *last_bay = DATAIDX(best).last_bay;
  fl_free(lp->data_fl, lst_look(lp->index,best));
  lst_del(lp->index,best);    
}

float
lp_best(lp_lst *lp)
{
	return (DATAIDX(0).v1);
}

lst
lp_bestlp(lp_lst *lp)
{
	return (DATAIDX(0).lp);
}

int
lp_bestlit(lp_lst *lp)
{
	return (DATAIDX(0).lit);
}

float
lp_worse(lp_lst *lp)
{
  int num;
  num = lst_size(lp->index);
  return (DATAIDX(num-1).v1);
}

int
lp_size(lp_lst *lp)
{
	return (lst_size(lp->index));
}

float
lp_bits(lp_lst *lp, int pos)
{
	return (DATAIDX(pos).v2);
}

float
lp_bay(lp_lst *lp, int pos)
{
	return (DATAIDX(pos).v1);
}

void
lp_free(lp_lst *lp,tab *t)
{
	int num, j, p, lit;

	if (lp->lev != 0) {
	  num = lst_size(lp->index);
	  for(j=0;j<num;j++) {
	    p = DATAIDX(j).lp;
	    lit = DATAIDX(j).lit;
	    if (p != -1) lp_tree_remove(p,lp->lev);
		if (lit != -1) fl_free(t->lit_free_list,lit);
	  }
	}
	mem_free(lp->data);
	lst_free(lp->index);
	fl_dispose(lp->data_fl);
	mem_free(lp);
}

int
lp_better_then_worse(lp_lst *lp, float h)
{
  int num;

  num = lst_size(lp->index);
  /*printf("btw num %d lps %d v1 %e h %e r %d \n",num,lp->size,(DATAIDX(num-1).v1),h,((DATAIDX(num-1).v1) < h));*/
  if (num < lp->size) return TRUE;
  return ((DATAIDX(num-1).v1) < h);
}

int
lp_add(lp_lst *lp, lst p, float v1, float v2, int lit, float theta, tab *t, float last_bay)
{
  int num, k, lit_bot, pout, d_loc;
  int isin;

  if(DB(dblp)) printf("lp_add %d %e %e %d %e %e\n",p,v1,v2,lit,theta,last_bay);
  
  num = lst_size(lp->index);
  
  if (num == 0) {
    d_loc = fl_get(lp->data_fl);
    lst_append(lp->index,d_loc);
    (*lp->data)[d_loc].lp = p;
    (*lp->data)[d_loc].lit = lit;
    (*lp->data)[d_loc].v1 = v1;
    (*lp->data)[d_loc].v2 = v2;
    (*lp->data)[d_loc].theta = theta;
    (*lp->data)[d_loc].last_bay = last_bay;
  } else {
    if (num < lp->size) {
      k=0;
      isin = FALSE;
      if (lp->lev != 0) {
	while ((k < num) && (v1 <= (DATAIDX(k).v1))) {
	  if ((v1 == (DATAIDX(k).v1)) && (lst_cmp(DATAIDX(k).lp,p))) isin = TRUE;
	  k++;
	}
      } else {
	while ((k < num) && (v1 <= (DATAIDX(k).v1))) {
	  k++;
	}
      }
      if (isin) {
	if (p != -1) lp_tree_remove(p,lp->lev);
	if (lit != -1) fl_free(t->lit_free_list,lit);
	return FALSE;
      } else {
	if (k == num) {
	  d_loc = fl_get(lp->data_fl);
	  lst_append(lp->index,d_loc);
	  (*lp->data)[d_loc].lp = p;
	  (*lp->data)[d_loc].lit = lit;
	  (*lp->data)[d_loc].v1 = v1;
	  (*lp->data)[d_loc].v2 = v2;
	  (*lp->data)[d_loc].theta = theta;
	  (*lp->data)[d_loc].last_bay = last_bay;
	} else {
	  d_loc = fl_get(lp->data_fl);
	  lst_shift(lp->index,d_loc,k);
	  (*lp->data)[d_loc].lp = p;
	  (*lp->data)[d_loc].lit = lit;
	  (*lp->data)[d_loc].v1 = v1;
	  (*lp->data)[d_loc].v2 = v2;
	  (*lp->data)[d_loc].theta = theta;
	  (*lp->data)[d_loc].last_bay = last_bay;
	}
      }
    } else {
      if (v1 <= (DATAIDX(num-1).v1)) {
	if (p != -1) lp_tree_remove(p,lp->lev);
	if (lit != -1) fl_free(t->lit_free_list,lit);
	return FALSE;
      } else {
	k=0;
	isin = FALSE;
	while ((k < num) && (v1 <= (DATAIDX(k).v1))) {
	  if ((v1 == (DATAIDX(k).v1)) && (lst_cmp(DATAIDX(k).lp,p))) 
	    isin = TRUE;
	  k++;
	}
	if (k == num) error ("k=num in form program");
	if (isin) {
	  if (p != -1) lp_tree_remove(p,lp->lev);
	  if (lit != -1) fl_free(t->lit_free_list,lit);
	  return FALSE;
	} else {
	  d_loc = lst_pop(lp->index);
	  if ((pout = (*lp->data)[d_loc].lp) != -1) lp_tree_remove(pout,lp->lev);
	  if ((lit_bot = (*lp->data)[d_loc].lit) != -1) fl_free(t->lit_free_list,lit_bot);
	  
	  lst_shift(lp->index,d_loc,k);
	  (*lp->data)[d_loc].lp = p;
	  (*lp->data)[d_loc].lit = lit;
	  (*lp->data)[d_loc].v1 = v1;
	  (*lp->data)[d_loc].v2 = v2;
	  (*lp->data)[d_loc].theta = theta;
	  (*lp->data)[d_loc].last_bay = last_bay;
	}
      }
    }
  }
  return TRUE;
}

/* This calculates the baysian estimation if this is the finial clause set. */
/* Uses the fixed example size with noise. */

float
cal_theta(tab *t, float tc)
{
  float theta,nt;

  nt = t->num_w_t;

  if (tc == 0.0) {
    theta = 0.5 / nt;
  } else if (tc == nt) {
    theta = (nt - 0.5) /nt;
  } else {
    theta = (tc*1.0) / nt;
  }

  theta = theta * 0.8 + 0.1;

  return theta;
}

float
bay(tab *t, float bits, float pc, float nc, float theta)
{
  float np, nn;
  float TP, TN, FP, FN;
  float r;

  np = t->num_w_p;
  nn = t->num_w_n;
  TP = pc;
  FP = np - pc;
  TN = nn - nc;
  FN = nc;
  switch(model) {
  case MOD_FIXED_NOISE :
    if (noise == 0.0) {
      if (FP + FN > 0.0) {
	r = BIGNEG + TP + TN;
      } else {
	r = TP * (log2(1/theta)) + TN * (log2(1/(1-theta))) + bits;
      }
    } else {
      r = TP * (log2((1.0-noise)/theta + noise)) +
        TN * (log2((1.0-noise)/(1-theta) + noise)) +
        (FP + FN) * (log2(noise)) + bits;
    }
    break;
  default : 
    r = 0.0;
    break;
  }
  /*printf ("bay b:%6.2e pc:%6.2f np:%6.2f nc:%6.2f nn:%6.2f t:%6.2f ", bits,pc,np,nc,nn,theta);
  printf ("=> r:%6.3f\n", r);*/  
  return r;
}

/* This calculates the best Baysian set of clause if other clauses are added to this set. */ 

float 
minf (float a, float b) {
  if (a < b) return a;
  return b;
}

float
best_bay(tab *t, float bits, float pc, float nc, float theta)
{
  float TP, TN, FP, FN, np, nn;
  float r, s;

 

  np = t->num_w_p;
  nn = t->num_w_n;

  TP = np;
  FP = 0.0;
  TN = nn - nc;
  FN = nc;
  if (pc != 0) {
    s = np/pc;
  } else {
    s = 1000000.0;
  }

  theta = minf (theta * (2-pow(2.0,-s)),1.0);
  theta = theta * 0.8 + 0.1;


  if (noise == 0.0) {
    if (FN > 0.0) {
      r = BIGNEG - FN;
    } else {
      r = TP * log2(1/theta) + TN * log2(1/(1-theta)) + bits*s;
    }
  } else {
    /*   if (((np*1.0)/((np + nn)*1.0) > 0.05) || (np > 49)) {*/
      r = TP * (log2((1.0-noise)/(theta) + noise)) +
	(nn-FN*s) * (log2((1.0-noise)/(1.0-theta) + noise)) +
	(FP + FN*s) * (log2(noise)) + (bits*s);
      /*  } else {
	r = TP * (log2((1.0-noise)/theta + noise)) +
          TN * (log2((1-noise)/((1-theta)/2) + noise)) +
	  (FP + FN) * (log2(noise)) + (bits*1.5);*/ /* could change 1.5 to -(bits/numclause) */
      /*}*/
  }

  /*  printf ("bb b:%6.2e pc:%6.2f np:%6.2f nc:%6.2f nn:%6.2f t:%6.2f ", bits,pc,np,nc,nn,theta);
  printf ("=> r:%6.3f\n", r);*/
  return r;
}

/* This calculates the best bayisian estimation
   if it is anded with other simple clauses. */

float
best_and_bay(tab *t, float bits, float pc, float nc, float theta)
{
  float TP, TN, FP, FN, np, nn;
  float r, linc, cinc;

  np = t->num_w_p;
  nn = t->num_w_n;

  TP = np;
  FP = 0.0;
  TN = nn;
  FN = 0.0;

  switch(model) {
  case MOD_FIXED_NOISE :
    if (nn == 0.0) {
      linc = 1.0;
    } else {
      linc = ((float) nn) / ((float) (nn - nc));
    }
    cinc = ((float) np) / ((float) pc);

    if (noise == 0.0) {
      if (np != 0.0) {
	r = TP * (log2(1.0/theta)) +
	  TN * (log2(1.0/(1.0-theta))) +
	  + (bits*linc*cinc);
      } else {
	r = TP * (log2(1.0/theta)) +
	  TN * (log2(1.0/(1.0-theta))) +
	  + bits;
      }     
    } else {
      if (np != 0.0) {
	r = TP * (log2((1.0-noise)/theta + noise)) +
	  TN * (log2((1.0-noise)/(1.0-theta) + noise)) +
	  (FP + FN) * (log2(noise)) + (bits*linc*linc*linc*cinc);
      } else {
	r = TP * (log2((1.0-noise)/theta + noise)) +
	  TN * (log2((1.0-noise)/(1.0-theta) + noise)) +
	  (FP + FN) * (log2(noise))  + bits;
      }
    }
    break;
  default : 
    r = 0.0;
    break;
  }

  /*  printf ("bab => noise:%1.3f t:%1.3f bits:%3.1f  TP:%3.1f TN:%3.1f FP:%3.1f FN:%3.1f => r:%3.3f\n", noise, t, bits, TP, TN, FP, FN,r);*/
  return r;
}


float
gain_h(float p, float n, float p_old, float n_old) {
  return ((n_old-n) * (log2 (p+2)));
}

float
gain_lp(float bits, float p, float n, float p_old, float n_old) {
  if (noise == 0.0) {
    return ((log2 (p+2)) * (2000.0*(n_old-n) + (p-p_old)) + bits * 0.01);
  } else {
    return ((log2 (p+2)) * (2.0*(n_old-n) + (p-p_old)) + bits * 0.01);
  } 
}





/* returns the log2 of the probability  of a logic program */
float
prob_lp (lst lp, info *i)
{
  float p;
  int s,j;

  s = lst_size(lp);
  p = log_geo(s,mu_clauses); 
  for (j=0;j<s;j++) {
    p = p + prob_clause(lst_look(lp,j),i);
  }
  p = p*apriorfactor; /* experiment with apoir consider bigger hyp */
  return (p);
}

float
prob_clause (lst c, info *i)
{
  float p;
  int s,j, invs, outvs, numlits;
 lst invsl, outvsl;
	
  numlits = ps_bg->num;
  s = lst_size(c);
  p = log_geo(s,mu_lit) - s*log2(numlits);
  for (j=0;j<s;j++) {
    invsl = (*i->lit_tab)[lst_look(c,j)].in_vars;
    if (invsl == -1) {
      invs = 0;
    } else {
      invs = lst_size (invsl);
    }
    outvsl = (*i->lit_tab)[lst_look(c,j)].out_vars;
    if (outvsl == -1) {
      outvs = 0;
    } else {
      outvs = lst_size (outvsl);
    }


    p = p - 2.3*(invs + outvs);
  }
  return p;
}

float
log_geo (int n, float mu)
{
  return (n*(log2(1-(1/mu))));
}


search
*init_search(int gain_list_size,
	     int max_clause_best, 
	     int max_clause_set_possible, 
	     int max_clause_set_final)
{
  search *ss;
  ss = (search *) memory (sizeof(search),"create search struct");
  ss->gain_list_size = gain_list_size;
  ss->clause_best =  lp_create(max_clause_best,1); 
  ss->gain_list_size_logic =  max_clause_set_possible; 
  ss->clause_set_final = lp_create(max_clause_set_final,2); 
  return ss;
}

void
add_simple_clause_to_pos (search *ss, tab *t, info *i)
{
  int litp, lit, j;
  float pc, nc, tc;
  lst tlu, lu;
  float bits, poss_r, theta;

  if (DB(dbsearchb) || DB(dbsearch)) printf("\n\n");
  if (DB(dbstage))  printf("==Setup Clauses\n");

  tab_hash_bits_clear(t); 

  for(j=0;j<i->num_lit_tab ;j++) {
    lit = (*i->lit_tab)[j].lit_pos;
    if (lit != -1) { 
                           /* copy to clause_best */
      litp = tab_replicate(t,lit);
      tab_cover(t,lit,&pc,&nc,&tc);
      tlu = (*i->lit_tab)[j].lits_used;
      lu = lst_replicate(tlu);
      bits = prob_clause(lu,i);
      theta = cal_theta(t,tc);

      poss_r = best_bay(t,bits, pc, nc, theta);
    
      if (poss_r > BIGNEG && ((tab_bits_hash_notfindadd(t, litp) ) != -1)) {
	lp_add(ss->clause_best,lu,poss_r,bits,litp,theta,t,0.0);
      } else {
	lst_free(lu);
	fl_free(t->lit_free_list,litp);
      }
    }  
  }  
}

int 
cal_gain_list_size(int rsize, int lev) {
  double levf;
  double rsizef;
  double power;
  double fres;

  if (lev == 0) {
    return rsize;
  } else {
    levf = lev;
    rsizef = rsize;
    power = 1/pow(2.0,levf);
    fres = pow(rsizef,power);
    if (fres <= 1.0) {
      return 1;
    } else {
      return fres;
    } 
  }
}

void
probe_clause(tab *t, info *i, search *ss, int c,lst cp, int lev, lst gain_order_ranking, int prevj)
{
  int k, j, w;
  lp_lst *gain_list;
  float pc1, nc1, tc1, pc2, nc2, tc2;
  int lit, cn, cnr;
  lst tlu, cpnr, cpn;
  float gain, theta, bits, poss_r;


  if (DB(dbsearchb)) {
    printf("---probe_clause lev %2d gl_size %3d jrank %2d clause ",lev,cal_gain_list_size(ss->gain_list_size,lev),prevj);
    show_c (stdout, cp, i);
    printf("\n");
  }
  if (lev <= BC_DEPTH) {
    tab_cover(t,c,&pc1,&nc1,&tc1);
    gain_list = lp_create(cal_gain_list_size(ss->gain_list_size,lev),1);
    for(j=prevj+1;j<i->num_lit_tab && lst_look(gain_order_ranking,j) != -1;j++) {
      k = lst_look(gain_order_ranking,j);
      lit = (*i->lit_tab)[k].lit_pos;
      if (lit != -1) {
    	tlu = (*i->lit_tab)[k].lits_used;
	/*printf("considering ");
	  show_c(stdout,tlu,i);*/
/*printf("p %3.1f n %3.1f t %3.1f => ",pc1,nc1,tc1);*/

	if (!lst_subset(tlu,cp)) {
	  
	  cn = fl_get(t->lit_free_list);
	  bvs_and(t->lit_cover_p,c,lit,cn);
	  bvs_and(t->lit_cover_n,c,lit,cn);
	  bvs_and(t->lit_cover_t,c,lit,cn);
	  cal_cover(t,cn);
	  tab_cover(t,cn,&pc2,&nc2,&tc2);
	  /*printf("p %3.1f n %3.1f t %3.1f => ",pc2,nc2,tc2);*/
	  if (((pc2 == 0.0) && (nc2 == 0.0) && (tc2 == 0.0)) || 
	      ((nc2 == 0.0) && (nc1 == 0.0) && (tc2 == tc1)) ||
              ((pc2 == 0.0) && (t->num_w_p > (t->num_w_n/20))) ||
              ((nc1 == nc2) && (tc1 == tc2))) {
	    fl_free(t->lit_free_list,cn);
	    /*printf(" free ");*/
	  } else {
	    gain = gain_h(pc2,nc2,pc1,nc1);
	    theta = cal_theta(t,tc2);
	    cpn = lst_new(lst_size(cp) + lst_size(tlu));
	    lst_merge(cp,tlu,cpn);
	    bits = prob_clause(cpn,i);
	    poss_r = best_bay(t,bits,pc2,nc2,theta);
	    if (poss_r > BIGNEG) {
	      cnr = tab_replicate(t,cn);
	      cpnr = lst_replicate(cpn);
	      if (((tab_bits_hash_notfindadd(t, cnr) ) != -1)) {
		/*printf(" add %3.1f ",poss_r);*/
		lp_add(ss->clause_best,cpnr,poss_r,bits,cnr,theta,t,0.0);
	      } else {
		lst_free(cpnr);
		fl_free(t->lit_free_list,cnr);
	      }
	    }
	    if ( (lst_size(gain_list->index) < gain_list->size) || 
		 (lp_worse(ss->clause_best) < best_and_bay(t,bits,pc2,nc2,theta))) {
	      lp_add(gain_list,cpn,gain,bits,cn,theta,t,j);
	      /*printf(" consider gain list ");*/
	    } else {
	      /*printf(" not consider gain list ");*/
	      fl_free(t->lit_free_list,cn);
	      lst_free(cpn);
	    }
	  } 
	  
	}
	/*printf("\n");*/
      }
    }
    if (lp_size(gain_list) > 0) {
      for (w=0;w<lp_size(gain_list);w++) {
	lit = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).lit;
	tlu = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).lp;
	j = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).last_bay;
	probe_clause(t,i,ss,lit,tlu,lev+1,gain_order_ranking,j);
      }
    }
    lp_free(gain_list,t);
  }
}

lst
get_gain_order_ranking(tab *t, info *i) {
  int k;
  lst res;
  lp_lst *gain_list;
  int w;
  float pc1, nc1, tc1, pc2, nc2, tc2;
  int lit;
  float gain;
  
  gain_list = lp_create(i->num_lit_tab,0);
  res = lst_new(i->num_lit_tab);
  lst_fill(res,i->num_lit_tab,-1);
  tab_cover(t,0,&pc1,&nc1,&tc1);
  for(k=0;k<i->num_lit_tab;k++) {
      lit = (*i->lit_tab)[k].lit_pos;
      if (lit != -1) {
	tab_cover(t,lit,&pc2,&nc2,&tc2);
	gain = gain_h(pc2,nc2,pc1,nc1);
	lp_add(gain_list,0,gain,0.0,k,0.0,t,0.0);
      }
  }
  for (w=0;w<lp_size(gain_list);w++) {
    k = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).lit;
    lst_insert(res,k,w);
  }
  if (DB(dbsearchb))  {
    printf("=== gain list for simple clauses.\n");
    lp_show(gain_list,t,i);
  }
  lp_free(gain_list,t);  
  return res;
}

void
form_clauses(search *ss, tab *t, info *i)
{
  lst c;
  lst gain_order_ranking;
  int lit, k, tlu;
  float pc2, nc2, tc2;

  /* combing literal sets to form clauses */

  if (DB(dbsearchb) || DB(dbsearch)) printf("\n\n");  
  if (DB(dbsearchb))  {
    printf("=== best simple clauses.\n");
     for(k=0;k<i->num_lit_tab;k++) {
      lit = (*i->lit_tab)[k].lit_pos;
      if (lit != -1) {
	tab_cover(t,lit,&pc2,&nc2,&tc2);
	tlu = (*i->lit_tab)[k].lits_used;
	printf("%4d : p %5.1f n %5.1f t %5.1f : ", lit, pc2, nc2, tc2);
	show_c(stdout,tlu,i);
	printf("\n");
      }
  } 
  }
  if (DB(dbstage)) printf ("\n==Simple Clauses -> Clauses\n");

  if (DB(dbsearchb)) {      
    printf("\n=== Clause Best List - before probe_clause\n");
    lp_show(ss->clause_best,t,i);
  }

  gain_order_ranking = get_gain_order_ranking(t,i);

  c = lst_new(10);
  probe_clause(t,i,ss,0,c,0,gain_order_ranking,-1);

  lst_free(gain_order_ranking);

 if (DB(dbsearchb)) {      
    printf("\n=== Clause Best List - after probe_clause\n");
    lp_show(ss->clause_best,t,i);

  }
}




void
probe_logic(tab *t, info *i, search *ss, int c,lst cp, int lev)
{
  int j, w, thisrecursive, allrecursive;
  lp_lst *gain_list;
  float pc1, nc1, tc1, pc2, nc2, tc2;
  int lit, cn, cnr;
  lst tlu, cpnr, cpn;
  float gain, theta, bits, r;
  int steps;

  if (DB(dbsearchb)) {
    printf("===probe_logic lev %2d gl_size %3d  ",lev,cal_gain_list_size(ss->gain_list_size_logic,lev));
    show_lp_nostats (stdout, cp, i);
    printf("\n");
  }
  if (lev <= BC_DEPTH) {
    tab_cover(t,c,&pc1,&nc1,&tc1);
    gain_list = lp_create(cal_gain_list_size(ss->gain_list_size_logic,lev),2);
    for(j=0;j<lp_size(ss->clause_best);j++) {
      tlu = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lp;
      if (!lst_member(tlu,cp)) {
	lit = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lit;
	cn = fl_get(t->lit_free_list);
	bvs_or(t->lit_cover_p,c,lit,cn);
	bvs_or(t->lit_cover_n,c,lit,cn);
	bvs_or(t->lit_cover_t,c,lit,cn);
	cal_cover(t,cn);
	tab_cover(t,cn,&pc2,&nc2,&tc2);
	theta = cal_theta(t,tc2);
	cpn = lst_new(lst_size(cp) + lst_size(tlu));
	lst_copy(cp,cpn);
	lst_append(cpn,tlu);
	if (DB(dbsearch)) {	
	  printf("consider \n");
	  show_lp_nostats(stdout,cpn,i);
	  printf("startat %2.4e %2.4e %2.4e => %2.4e %2.4e %2.4e \n", pc1, nc1, tc1, pc2, nc2, tc2);
	}
	bits = prob_lp(cpn,i);
	gain = gain_lp(bits,pc2,nc2,pc1,nc1);
	r = best_bay(t,bits,pc2,nc2,theta);

	/*printf("gain : %2.4e r : %2.4e lpbest : %2.4e ", gain, r, lp_best(ss->clause_set_final));*/

	if (lev == 0 && no_rec_lits_lp(cpn,i)) gain += 10000.0; /* help start off with non-recursive clause */

	thisrecursive = recursive && !no_rec_lits_lp(cpn,i);
	allrecursive = all_rec_lits_lp(cpn,i);
	
	cnr = tab_replicate(t,cn);
	cpnr = lst_replicate(cpn);
	
	/*printf("gain : %2.4e ",gain);*/
	if ((gain > 0.0) && (lp_size(ss->clause_set_final) == 0 ||(r > lp_best(ss->clause_set_final)))) {
	 
	  /*printf("added to gain list\n");*/
	  lp_add(gain_list,cpn,gain,bits,cn,theta,t,lp_bits(ss->clause_best,j));
	} else {
	  /*printf("NOT added to gain list\n");*/
	  lst_free(cpn);
	  fl_free(t->lit_free_list,cn);
	}
	r = bay(t, bits, pc2, nc2, theta);
	if (!thisrecursive) {
	  if (DB(dbsearchb)) {
	      if (lp_better_then_worse(ss->clause_set_final,r)) {
		printf(" Is better.pc:%6.2f nc:%6.2f q:%14.2f\n",pc2,nc2,r);
		show_lp_nostats (stdout,cpnr,i);
		printf("\n");
	      } /* else {
		printf(" Is not better.pc:%6.2f nc:%6.2f q:%14.2f\n",pc2,nc2,r);
		show_lp_nostats (stdout,cpnr,i);
		printf("\n");
		} 	*/    
	  }
	  lp_add(ss->clause_set_final,cpnr,r,bits,cnr,theta,t,0.0);
	} else {
	  /*printf("before %2.4e %2.4e %2.4e %2.4e => r %2.4e \n",bits,pc2,nc2,theta,r);*/
	  /*printf ("\n==== Is better than woarst ??? . %6.2f %6.2f ,  %2.4e > %2.4e => \n",
	    pc2,nc2,r,lp_worse(ss->clause_set_final));*/
	  if(lp_better_then_worse(ss->clause_set_final,r)) {
	   
	    if (DB(dbsearchb)) {
	      printf ("\n==== Consider for resolution. %6.2f %6.2f %2.4e > %2.4e => ",
		      pc2,nc2,r,lp_worse(ss->clause_set_final));
	    }
	    if (allrecursive) {
	      no_cover(t,cnr);
	      if (DB(dbsearchb)) {printf("(all rec) ");}
	    } else {
	      cover(i,cpnr,&steps,cnr);
	    }
	    
	    tab_cover(t,cnr,&pc2,&nc2,&tc2);
	    theta = cal_theta(t,tc2);
	    
	    r = bay(t, bits, pc2, nc2, theta);
	    /*  printf("after %2.4e %2.4e %2.4e %2.4e => r %2.4e \n",bits,pc2,nc2,theta,r);*/
    
	    if (DB(dbsearchb)) {
	      if (lp_better_then_worse(ss->clause_set_final,r)) {
		printf(" Is better.pc:%6.2f nc:%6.2f q:%2.4e\n",pc2,nc2,r);
	      } else {
		printf(" Not better.pc:%6.2f nc:%6.2f q:%2.4e\n",pc2,nc2,r);
	      }
	      show_lp_nostats (stdout,cpnr,i);
	      printf("\n");
	    }
	      
	    lp_add(ss->clause_set_final,cpnr,r,bits,cnr,theta,t,0.0);
	  } else {
	    lst_free(cpnr);
	    fl_free(t->lit_free_list,cnr);
	  }
	  
	} 
      }
    }
  
    if (lp_size(gain_list) > 0) {
      /*printf("\ngain list "); lp_show(gain_list,t,i);*/
      for (w=0;w<lp_size(gain_list);w++) {
	lit = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).lit;
	tlu = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).lp;
	probe_logic(t,i,ss,lit,tlu,lev+1);
      }
    }
    lp_free(gain_list,t);
  }
}


void
form_programs(search *ss,tab *t,info *i)
{
  lst lp;
  int lit;
  float pc2,nc2,tc2;
  int j, tlu;

  /* combing literal sets to form clauses */

  if (DB(dbsearchb) || DB(dbsearch)) printf("\n\n");  
  if (DB(dbsearchb)) {
    printf ("=== Best clauses.\n");
    for(j=0;j<lp_size(ss->clause_best);j++) {
      tlu = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lp;
      lit = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lit;
      tab_cover(t,lit,&pc2,&nc2,&tc2);
      printf("%4d : p %5.1f n %5.1f t %5.1f : ", lit, pc2, nc2, tc2);
      show_c(stdout,tlu,i);
      printf("\n");
    }
  }

  if (DB(dbstage)) printf ("==Clauses --> Logic Program\n");

  lp = lst_new(10);
  lit = fl_get(t->lit_free_list);
  tab_clear(t,lit);
  probe_logic(t,i,ss,lit,lp,0);

 if (DB(dbsearchb)) {      
    printf("\n=== Best Logic Programs\n");
    lp_show(ss->clause_set_final,t,i);

  }
}


void
join_programs(search *ss,tab *t,info *i)
{
  int j;
  int litp, lit, lpos, lpostemp;
  float pc, nc, tc;
  float pc1, nc1, tc1;
  float pc2, nc2, tc2;
  lst pn, pn2, tlu;
  float bits, r,  theta;


  lpos = fl_get(t->lit_free_list);      /* add to clause set possible */
  lpostemp = fl_get(t->lit_free_list); 
  tab_clear(t,lpos);
  tab_clear(t,lpostemp);
  pn = lst_new(500);

  for (j=0;j<lp_size(ss->clause_best);j++) {
    if (DB(dbsearchb)) printf("====Start join for loop\n");
    tlu = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lp;
    lit = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lit;
	
   
    bvs_or(t->lit_cover_p,lit,lpos,lpostemp);
    bvs_or(t->lit_cover_n,lit,lpos,lpostemp);
    bvs_or(t->lit_cover_t,lit,lpos,lpostemp);
    cal_cover(t,lpostemp);
    tab_cover(t,lpostemp,&pc,&nc,&tc);

    tab_cover(t,lit,&pc1,&nc1,&tc1);
    tab_cover(t,lpos,&pc2,&nc2,&tc2);

    if (((nc - nc2) < 5.0 && pc > pc2) ||
        ((pc-pc2)/(nc-nc2) > 2.0)) {

      tab_copy(t,lpostemp,lpos);
      lst_append(pn,tlu);
      
      if (DB(dbsearchb)) {
	printf("===The joint hyppothesis so far\n");
	show_lp (stdout,pn,i);
	printf("\n");    
      }
      
      
      litp = tab_replicate(t,lpos);
      pn2 = lst_replicate(pn);
      bits = 10.0; /*prob_lp(pn2,i);*/
      theta = cal_theta(t,tc);
      r = bay(t, bits, pc, nc, theta);
      r = j;
      lp_add(ss->clause_set_final,pn2,r,bits,litp,theta,t,0.0);
    } 
  }
  fl_free(t->lit_free_list,lpos); 
  fl_free(t->lit_free_list,lpostemp);
}


void
do_search(info *i,search *ss)
{
tab *t;

  t = i->the_tab;
	
  add_simple_clause_to_pos (ss,t,i);
  form_clauses (ss,t,i);
  invent_predicate(ss,t,i);
  form_programs(ss,t,i); 
  /* join_programs(ss,t,i);*/
}








