/* search.c */

#include "main.h"


#define NUM_LP 5
#define DATAIDX(X) ((*lp->data)[lst_look(lp->index,X)])

#define LOOP_BOUND_C 1000
#define LOOP_BOUND_P 1000


lp_lst
*lp_create(int size, int lev)
{
	lp_lst *lp;

	lp = (lp_lst *) memory(sizeof(lp_lst),"creating search block");
        lp->index = lst_new(size);
	lp->data = (lp_data (*)[]) memory (sizeof(lp_data)*size,"data for lp");
        lp->data_fl = fl_create(size);
	lp->size = size;
	lp->lev = lev;
	return (lp);
}

void
lp_reset(lp_lst *lp, tab *t) 
{
  int num, j, p, lit;
  
  num = lst_size(lp->index);
  for(j=0;j<num;j++) {
    p = DATAIDX(j).lp;
    lit = DATAIDX(j).lit;
    if (p != -1) lp_tree_remove(p,lp->lev);
    if (lit != -1) fl_free(t->lit_free_list,lit);
  } 
  lst_clear(lp->index);
  fl_clear(lp->data_fl);
}

void
lp_show_j (lp_lst *lp, tab *t, info *i, int j)
{
  float pc, nc, tc;
  
  tab_cover(t,(*lp->data)[lst_look(lp->index,j)].lit,&pc,&nc,&tc);
  printf("%4d  %6d  %6d  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f/%6.2f  %6.2f/%6.2f  ",
         j,	 
	 DATAIDX(j).lp,
	 DATAIDX(j).lit,
	 DATAIDX(j).v1,
	 DATAIDX(j).v2,
	 DATAIDX(j).theta,
	 DATAIDX(j).last_bay,
	 pc,
	 t->num_w_p,
	 nc,
	 t->num_w_n);
  if (lp->lev == 2) {
    printf("\n");
    show_lp (stdout, DATAIDX(j).lp, i);
  } else if (lp->lev == 1) {
    show_c (stdout, DATAIDX(j).lp, i);
    printf("\n");
 } 
}

void
lp_show (lp_lst *lp, tab *t, info *i)
{
  int j, num;

  printf("size %d lev %d\n",lp->size,lp->lev);
  printf("   j      lp     lit      v1       v2   theta   last    pc/np   nc/nn\n");
  num = lst_size(lp->index);
    for(j=0;j<lst_size(lp->index);j++) {
      lp_show_j (lp,t,i,j);
    }
 
}

void
lp_tree_remove (lst idx, int lev)
{
	int i;

	if (lev == 2 || lev == 1)
		lst_free(idx);
	if (lev > 2) {
		for (i=0;i<lst_size(idx);i++)
			lp_tree_remove(lst_look(idx,i),lev-1);
		lst_free(idx);
	}
}

void
lp_copy(lst s, lst d)
{
	int j, num;
	lst c, k;

	lst_clear(d);
	num = lst_size(s);
	for (j=0;j<num;j++) {
		c = lst_look(s,j);
		k = lst_new(lst_size(c));
		lst_copy(c,k);
		lst_append(d,k);
	}

}

void
lp_pop(lp_lst *lp, tab *t) 
{
  lst p;
  int lit;

  p = DATAIDX(0).lp;
  lit = DATAIDX(0).lit;
  if (p != -1) lp_tree_remove(p,lp->lev);
  if (lit != -1) fl_free(t->lit_free_list,lit);  
  fl_free(lp->data_fl, lst_look(lp->index,0));
  lst_head(lp->index);   
}

void
lp_head(lp_lst *lp, lst *p, float *bits, int *lit, float *last_bay)
{
	*lit = DATAIDX(0).lit;
        *p = DATAIDX(0).lp;
	*bits = DATAIDX(0).v2;
	*last_bay = DATAIDX(0).last_bay;
        fl_free(lp->data_fl, lst_look(lp->index,0));
	lst_head(lp->index);    
}

void
lp_goodsmall(lp_lst *lp, float bestf, lst *p, float *bits, int *lit, float *last_bay)
{
  int k, best, size;
  float bestv;

  size = lp_size(lp);
  best = 0;
  bestv = DATAIDX(best).v2;
  k = 0;
  while ((k < (size-1)) && (DATAIDX(k).v1 > bestf)) {
    k++;
    if (DATAIDX(k).v2 < bestv) {
	bestv = DATAIDX(k).v2;
	best = k;
    }
  }
  
  /*printf ("k %d size %d best %d bestv %e\n",k,size,best,bestv);*/
  *lit = DATAIDX(best).lit;
  *p = DATAIDX(best).lp;
  *bits = DATAIDX(best).v2;
  *last_bay = DATAIDX(best).last_bay;
  fl_free(lp->data_fl, lst_look(lp->index,best));
  lst_del(lp->index,best);    
}

float
lp_best(lp_lst *lp)
{
	return (DATAIDX(0).v1);
}

lst
lp_bestlp(lp_lst *lp)
{
	return (DATAIDX(0).lp);
}

int
lp_bestlit(lp_lst *lp)
{
	return (DATAIDX(0).lit);
}

float
lp_worse(lp_lst *lp)
{
  int num;
  num = lst_size(lp->index);
  return (DATAIDX(num-1).v1);
}

int
lp_size(lp_lst *lp)
{
	return (lst_size(lp->index));
}

float
lp_bits(lp_lst *lp, int pos)
{
	return (DATAIDX(pos).v2);
}

float
lp_bay(lp_lst *lp, int pos)
{
	return (DATAIDX(pos).v1);
}

void
lp_free(lp_lst *lp,tab *t)
{
	int num, j, p, lit;

	num = lst_size(lp->index);
	for(j=0;j<num;j++) {
		p = DATAIDX(j).lp;
		lit = DATAIDX(j).lit;
		if (p != -1) lp_tree_remove(p,lp->lev);
		if (lit != -1) fl_free(t->lit_free_list,lit);
	}
	mem_free(lp->data);
	lst_free(lp->index);
	fl_dispose(lp->data_fl);
}

int
lp_better_then_worse(lp_lst *lp, float h)
{
  int num;

  num = lst_size(lp->index);
  /*printf("num %d lps %d v1 %e h %e r %d \n",num,lp->size,(*lp->v1)[num-1],h,(((*lp->v1)[num-1]) > h));*/
  if (num < lp->size) return TRUE;
  return ((DATAIDX(num-1).v1) < h);
}

int
lp_add(lp_lst *lp, lst p, float v1, float v2, int lit, float theta, tab *t, float last_bay)
{
  int num, k, lit_bot, pout, d_loc;
  int isin;

  if(DB(dblp)) printf("lp_add %d %e %e %d %e %e\n",p,v1,v2,lit,theta,last_bay);
  
  num = lst_size(lp->index);
  
  if (num == 0) {
    d_loc = fl_get(lp->data_fl);
    lst_append(lp->index,d_loc);
    (*lp->data)[d_loc].lp = p;
    (*lp->data)[d_loc].lit = lit;
    (*lp->data)[d_loc].v1 = v1;
    (*lp->data)[d_loc].v2 = v2;
    (*lp->data)[d_loc].theta = theta;
    (*lp->data)[d_loc].last_bay = last_bay;
  } else {
    if (num < lp->size) {
      k=0;
      isin = FALSE;
      while ((k < num) && (v1 <= (DATAIDX(k).v1))) {
	if ((v1 == (DATAIDX(k).v1)) && (lst_cmp(DATAIDX(k).lp,p))) isin = TRUE;
	k++;
      }
      if (isin) {
	if (p != -1) lp_tree_remove(p,lp->lev);
	if (lit != -1) fl_free(t->lit_free_list,lit);
	return FALSE;
      } else {
	if (k == num) {
	  d_loc = fl_get(lp->data_fl);
	  lst_append(lp->index,d_loc);
	  (*lp->data)[d_loc].lp = p;
	  (*lp->data)[d_loc].lit = lit;
	  (*lp->data)[d_loc].v1 = v1;
	  (*lp->data)[d_loc].v2 = v2;
	  (*lp->data)[d_loc].theta = theta;
	  (*lp->data)[d_loc].last_bay = last_bay;
	} else {
	  d_loc = fl_get(lp->data_fl);
	  lst_shift(lp->index,d_loc,k);
	  (*lp->data)[d_loc].lp = p;
	  (*lp->data)[d_loc].lit = lit;
	  (*lp->data)[d_loc].v1 = v1;
	  (*lp->data)[d_loc].v2 = v2;
	  (*lp->data)[d_loc].theta = theta;
	  (*lp->data)[d_loc].last_bay = last_bay;
	}
      }
    } else {
      if (v1 <= (DATAIDX(num-1).v1)) {
	if (p != -1) lp_tree_remove(p,lp->lev);
	if (lit != -1) fl_free(t->lit_free_list,lit);
	return FALSE;
      } else {
	k=0;
	isin = FALSE;
	while ((k < num) && (v1 <= (DATAIDX(k).v1))) {
	  if ((v1 == (DATAIDX(k).v1)) && (lst_cmp(DATAIDX(k).lp,p))) 
	    isin = TRUE;
	  k++;
	}
	if (k == num) error ("k=num in form program");
	if (isin) {
	  if (p != -1) lp_tree_remove(p,lp->lev);
	  if (lit != -1) fl_free(t->lit_free_list,lit);
	  return FALSE;
	} else {
	  d_loc = lst_pop(lp->index);
	  if ((pout = (*lp->data)[d_loc].lp) != -1) lp_tree_remove(pout,lp->lev);
	  if ((lit_bot = (*lp->data)[d_loc].lit) != -1) fl_free(t->lit_free_list,lit_bot);
	  
	  lst_shift(lp->index,d_loc,k);
	  (*lp->data)[d_loc].lp = p;
	  (*lp->data)[d_loc].lit = lit;
	  (*lp->data)[d_loc].v1 = v1;
	  (*lp->data)[d_loc].v2 = v2;
	  (*lp->data)[d_loc].theta = theta;
	  (*lp->data)[d_loc].last_bay = last_bay;
	}
      }
    }
  }
  return TRUE;
}

/* This calculates the baysian estimation if this is the finial clause set. */
/* Uses the fixed example size with noise. */

float
cal_theta(tab *t, float tc)
{
  float theta,nt;

  nt = t->num_w_t;

  if (tc == 0.0) {
    theta = 0.5 / nt;
  } else if (tc == nt) {
    theta = (nt - 0.5) /nt;
  } else {
    theta = (tc*1.0) / nt;
  }

  theta = theta * 0.8 + 0.1;

  return theta;
}

float
bay(tab *t, float bits, float pc, float nc, float theta)
{
  float np, nn;
  float TP, TN, FP, FN;
  float r;

  np = t->num_w_p;
  nn = t->num_w_n;
  TP = pc;
  FP = np - pc;
  TN = nn - nc;
  FN = nc;
  switch(model) {
  case MOD_FIXED_NOISE :
    if (noise == 0.0) {
      if (FP + FN > 0.0) {
	r = BIGNEG;
      } else {
	r = TP * (log2(1/theta)) + TN * (log2(1/(1-theta))) + bits;
      }
    } else {
      r = TP * (log2((1.0-noise)/theta + noise)) +
        TN * (log2((1.0-noise)/(1-theta) + noise)) +
        (FP + FN) * (log2(noise)) + bits;
    }
    break;
  default : 
    r = 0.0;
    break;
  }
  /*printf ("bay b:%6.2e pc:%6.2f np:%6.2f nc:%6.2f nn:%6.2f t:%6.2f ", bits,pc,np,nc,nn,theta);
  printf ("=> r:%6.3f\n", r);*/  
  return r;
}

/* This calculates the best Baysian set of clause if other clauses are added to this set. */ 

float
best_bay(tab *t, float bits, float pc, float nc, float theta)
{
  float TP, TN, FP, FN, np, nn;
  float r, s;

 

  np = t->num_w_p;
  nn = t->num_w_n;

  TP = (np+pc)/2.0;
  s = 0.0;
  TP = np;
  FP = 0.0;
  TN = nn - nc;
  FN = nc;
  if (pc != 0) {
    s = (1.0*np)/pc;
  } else {
    s = 1000000.0;
  }
  switch(model) {
  case MOD_FIXED_NOISE :
    if (noise == 0.0) {
      if (FN > 0.0) {
	r = BIGNEG;
      } else {
	r = TP * log2(1/theta) + TN * log2(1/(1-theta)) + bits*s;
      }
    } else {
      if (((np*1.0)/((np + nn)*1.0) > 0.05) || (np > 50)) {
	r = TP * (log2((1.0-noise)/(theta) + noise)) +
          (nn-FN*s) * (log2((1.0-noise)/(1.0-theta) + noise)) +
	  (FP + FN*s) * (log2(noise)) + (bits*s);
      } else {
	r = TP * (log2((1.0-noise)/theta + noise)) +
          TN * (log2((1-noise)/((1-theta)/2) + noise)) +
	  (FP + FN) * (log2(noise)) + (bits*1.5); /* could change 1.5 to -(bits/numclause) */
      }
    }
    break;  
  default : 
    r = 0.0;
    break;
  }
  /*  printf ("bb b:%6.2e pc:%6.2f np:%6.2f nc:%6.2f nn:%6.2f t:%6.2f ", bits,pc,np,nc,nn,theta);
  printf ("=> r:%6.3f\n", r);*/
  return r;
}

/* This calculates the best bayisian estimation
   if it is anded with other simple clauses. */

float
best_and_bay(tab *t, float bits, float pc, float nc, float theta)
{
  float TP, TN, FP, FN, np, nn;
  float r, linc, cinc;

  np = t->num_w_p;
  nn = t->num_w_n;

  TP = np;
  FP = 0.0;
  TN = nn;
  FN = 0.0;

  switch(model) {
  case MOD_FIXED_NOISE :
    if (nn == 0.0) {
      linc = 1.0;
    } else {
      linc = ((float) nn) / ((float) (nn - nc));
    }
    cinc = ((float) np) / ((float) pc);

    if (noise == 0.0) {
      if (np != 0.0) {
	r = TP * (log2(1.0/theta)) +
	  TN * (log2(1.0/(1.0-theta))) +
	  + (bits*linc*cinc);
      } else {
	r = TP * (log2(1.0/theta)) +
	  TN * (log2(1.0/(1.0-theta))) +
	  + bits;
      }     
    } else {
      if (np != 0.0) {
	r = TP * (log2((1.0-noise)/theta + noise)) +
	  TN * (log2((1.0-noise)/(1.0-theta) + noise)) +
	  (FP + FN) * (log2(noise)) + (bits*linc*linc*linc*cinc);
      } else {
	r = TP * (log2((1.0-noise)/theta + noise)) +
	  TN * (log2((1.0-noise)/(1.0-theta) + noise)) +
	  (FP + FN) * (log2(noise))  + bits;
      }
    }
    break;
  default : 
    r = 0.0;
    break;
  }

  /*  printf ("bab => noise:%1.3f t:%1.3f bits:%3.1f  TP:%3.1f TN:%3.1f FP:%3.1f FN:%3.1f => r:%3.3f\n", noise, t, bits, TP, TN, FP, FN,r);*/
  return r;
}


float
gain_h(float p, float n, float p_old, float n_old) {
  return ((n_old-n) * (log2 (p+2)));
}

float
gain_lp(float bits, float p, float n, float p_old, float n_old) {
  return ((log2 (p+2)) * (2.0*(n_old-n) + (p-p_old)) + bits * 0.01);
}





/* returns the log2 of the probability  of a logic program */
float
prob_lp (lst lp, info *i)
{
  float p;
  int s,j;

  s = lst_size(lp);
  p = log_geo(s,mu_clauses);
  for (j=0;j<s;j++) {
    p = p + prob_clause(lst_look(lp,j),i);
  }
  return (p);
}

float
prob_clause (lst c, info *i)
{
  float p;
  int s,j, invs, outvs, numlits;
 lst invsl, outvsl;
	
  numlits = ps_bg->num;
  s = lst_size(c);
  p = log_geo(s,mu_lit) - s*log2(numlits);
  for (j=0;j<s;j++) {
    invsl = (*i->lit_tab)[lst_look(c,j)].in_vars;
    if (invsl == -1) {
      invs = 0;
    } else {
      invs = lst_size (invsl);
    }
    outvsl = (*i->lit_tab)[lst_look(c,j)].out_vars;
    if (outvsl == -1) {
      outvs = 0;
    } else {
      outvs = lst_size (outvsl);
    }


    p = p - 2.3*(invs + outvs);
  }
  return p;
}

float
log_geo (int n, float mu)
{
  return (n*(log2(1-(1/mu))));
}


search
*init_search(int gain_list_size,
	     int max_clause_best, 
	     int max_clause_set_possible, 
	     int max_clause_set_final)
{
  search *ss;
  ss = (search *) memory (sizeof(search),"create search struct");
  ss->gain_list_size = gain_list_size;
  ss->clause_best =  lp_create(max_clause_best,1); 
  ss->clause_set_possible =  lp_create(max_clause_set_possible,2); 
  ss->clause_set_final = lp_create(max_clause_set_final,2); 
  return ss;
}

void
add_simple_clause_to_pos (search *ss, tab *t, info *i)
{
  int litp, lit;
  float pc, nc, tc;
  lst tlu, lu;
  float bits, poss_r, theta;

  if (DB(dbsearchb) || DB(dbsearch)) printf("\n\n");
  if (DB(dbstage))  printf("==Setup Clauses\n");


  lit = (*i->lit_tab)[0].lit_pos;
  if (lit != -1) { 
                           /* copy to clause_best */
    litp = tab_replicate(t,lit);
    tab_cover(t,lit,&pc,&nc,&tc);
    tlu = (*i->lit_tab)[0].lits_used;
    lu = lst_replicate(tlu);
    bits = prob_clause(lu,i);
    theta = cal_theta(t,tc);

    poss_r = best_bay(t,bits, pc, nc, theta);
    
    if (poss_r > BIGNEG) {
      lp_add(ss->clause_best,lu,poss_r,bits,litp,theta,t,0.0);
    } else {
      lst_free(lu);
      fl_free(t->lit_free_list,litp);
    }
  }    
}

int 
cal_gain_list_size(int rsize, int lev) {
  double levf;
  double rsizef;
  double power;
  double fres;

  if (lev == 0) {
    return rsize;
  } else {
    levf = lev;
    rsizef = rsize;
    power = 1/pow(2.0,levf);
    fres = pow(rsizef,power);
    if (fres <= 1.0) {
      return 1;
    } else {
      return fres;
    } 
  }
}

void
probe_clause(tab *t, info *i, search *ss,int j, int c,lst cp, int lev)
{
  int k, w;
  lp_lst *gain_list;
  float pc1, nc1, tc1, pc2, nc2, tc2;
  int lit, cn, cnr;
  lst tlu, cpnr, cpn;
  float gain, theta, bits, poss_r;

  if (DB(dbsearchb)) {
    printf("===probe_clause j %2d lev %2d gl_size %3d clause ",j,lev,cal_gain_list_size(ss->gain_list_size,lev));
    show_c (stdout, cp, i);
    printf("\n");
  }
  if (lev <= BC_DEPTH) {
    tab_cover(t,c,&pc1,&nc1,&tc1);
    gain_list = lp_create(cal_gain_list_size(ss->gain_list_size,lev),1);
    for(k=1;k<i->num_lit_tab;k++) {
      lit = (*i->lit_tab)[k].lit_pos;
      if (lit != -1) {
    	tlu = (*i->lit_tab)[k].lits_used;
	if (!lst_subset(tlu,cp)) {
	  
	  cn = fl_get(t->lit_free_list);
	  bvs_and(t->lit_cover_p,c,lit,cn);
	  bvs_and(t->lit_cover_n,c,lit,cn);
	  bvs_and(t->lit_cover_t,c,lit,cn);
	  cal_cover(t,cn);
	  tab_cover(t,cn,&pc2,&nc2,&tc2);
	  if (((pc2 == 0.0) && (nc2 == 0.0) && (tc2 == 0.0)) || 
	      ((nc2 == 0.0) && (nc1 == 0.0) && (tc2 == tc1)) ||
              ((pc2 == 0.0) && (t->num_w_p > (t->num_w_n/20))) ||
              ((nc1 == nc2) && (tc1 == tc2))) {
	    fl_free(t->lit_free_list,cn);
	  } else {
	    gain = gain_h(pc2,nc2,pc1,nc1);
	    theta = cal_theta(t,tc2);
	    cpn = lst_new(lst_size(cp) + lst_size(tlu));
	    lst_merge(cp,tlu,cpn);
	    bits = prob_clause(cpn,i);
	    poss_r = best_bay(t,bits,pc2,nc2,theta);
	    if (lev >= j && poss_r > BIGNEG) {
	      cnr = tab_replicate(t,cn);
	      cpnr = lst_replicate(cpn);
	      lp_add(ss->clause_best,cpnr,poss_r,bits,cnr,theta,t,0.0);
	    }
	    if ( (lst_size(gain_list->index) < gain_list->size) || 
		 (lp_worse(ss->clause_best) < best_and_bay(t,bits,pc2,nc2,theta))) {
	      lp_add(gain_list,cpn,gain,bits,cn,theta,t,0.0);
	    } else {
	      fl_free(t->lit_free_list,cn);
	      lst_free(cpn);
	    }
	  } 
	}
      }
    }
    if (lp_size(gain_list) > 0) {
      if (lev > j) {
	probe_clause(t,i,ss,j,
		     lp_bestlit(gain_list),
		     lp_bestlp(gain_list),lev+1);
      } else if (lev == j) {
	lp_pop(gain_list,t);
	for (w=0;w<lp_size(gain_list);w++) {
	  lit = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).lit;
	  tlu = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).lp;
	  probe_clause(t,i,ss,j,lit,tlu,lev+1);
	}
      } else {
	for (w=0;w<lp_size(gain_list);w++) {
	  lit = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).lit;
	  tlu = ((*(gain_list)->data)[lst_look((gain_list)->index,w)]).lp;
	  probe_clause(t,i,ss,j,lit,tlu,lev+1);
	}
      }
    }
    lp_free(gain_list,t);
  }
}

void
form_clauses(search *ss, tab *t, info *i)
{
  int j;
  lst c;

  /* combing literal sets to form clauses */

  if (DB(dbsearchb) || DB(dbsearch)) printf("\n\n");  
  if (DB(dbstage)) printf ("==Simple Clauses -> Clauses\n");

  j = -1;
  c = lst_new(10);
  while (j<FC_DEPTH) {
    lst_clear(c);
    probe_clause(t,i,ss,j,0,c,0);
    j++;
  }
 if (DB(dbsearchb)) {      
    printf("\n--- Clause Best List\n");
    lp_show(ss->clause_best,t,i);

  }
}

void
add_clause_to_pos (search *ss,tab *t,info *i)
{
  int j;
  int litp, lit, litp2;
  float pc, nc, tc; 
  int steps;
  lst p, tlu;
  float bits, r, poss_r, theta;

  if (DB(dbsearchb) || DB(dbsearch)) printf("\n\n");
  if (DB(dbstage)) printf ("==Setup Clause Sets\n");

  /* cimbining clause to form final logic programs */
  /* Creating the logic program lists */
  
  /* in1itialising the possible and adding to the final */
  for (j=0;j<lp_size(ss->clause_best);j++) {
    lit = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lit;
    litp = tab_replicate(t,lit);
    cal_cover(t,litp);
    tab_cover(t,litp,&pc,&nc,&tc);
    tlu = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lp;
    p = lst_new(1);
    lst_append(p,tlu);
    bits = prob_lp(p,i);
    theta = cal_theta(t,tc);
    poss_r = gain_lp(bits,pc,nc,0,0); 
    /*printf("@@@@@@@@@@@@ heur %e\n",poss_r);
      printf("b:%e pc:%6.2f np:%6.2f nc:%6.2f nn:%6.2f tc:%6.2f nt:%6.2f t:%e",bits,
      pc,t->num_w_p,
      nc,t->num_w_n,
      tc,t->num_w_t,
      theta);
      show_lp (stdout,p,i);*/

    lp_add(ss->clause_set_possible,p,poss_r,bits,litp,theta,t,lp_bits(ss->clause_best,j));
    
    
    /* add to final list */
    
    p = lst_new(1);
    lst_append(p,tlu);
    litp2 = tab_replicate(t,litp);
    cal_cover(t,litp2);    
    tab_cover(t,litp2,&pc,&nc,&tc);
    theta = cal_theta(t,tc);
    r = bay(t,bits,pc,nc,theta);
    if (!recursive) {
      lp_add(ss->clause_set_final,p,r,bits,litp2,theta,t,0.0); 
    } else {
      if (lp_better_then_worse(ss->clause_set_final,r)) {
	if (DB(dbsearch)) {
	  printf ("\n==== Consider for resolution. %2.4e ",r);
	}
	cover(i,p,&steps,litp2);
	tab_cover(t,litp2,&pc,&nc,&tc);
	theta = cal_theta(t,tc);
	r = bay(t, bits, pc, nc, theta);
	if (DB(dbsearch)) {
	  if (lp_better_then_worse(ss->clause_set_final,r)) {
	    printf(" Is better. %2.4e\n",r);
	  } else {
	    printf(" Not better.%2.4e\n",r);
	  }
	  show_lp (stdout,p,i);
	}
	lp_add(ss->clause_set_final,p,r,bits,litp2,theta,t,0.0); 
      } else {
	lst_free(p);
	fl_free(t->lit_free_list,litp2);
      }
    }
  }
  
  if (DB(dbsearchb)) {
    printf("\n~~~ Clause Set Possible List (used as the seed to logic programs)\n");
    lp_show(ss->clause_set_possible,t,i);
	  
    printf("\n~~~ Clause Set Final List  (the best so far)\n");
    lp_show(ss->clause_set_final,t,i);
  }
}

void
form_programs(search *ss,tab *t,info *i)
{
  int j;
  int litp, lit, lpos, litp2;
  float pc, nc, tc;
  float pc1, nc1, tc1;
  float pcold, ncold, tcold;
  lst p, pn, pn2, tlu;
  float bits, oldbits, r, poss_r, theta,last_bay, v1;
  int steps;
  int loopcount;

  loopcount = 0;

  if (DB(dbsearchb) || DB(dbsearch)) printf("\n\n");
  if (DB(dbstage)) printf("=== Combing Clauses\n");

  while ((lp_size(ss->clause_set_possible) > 0) &&
	 (loopcount < LOOP_BOUND_P) &&
         (lp_best(ss->clause_set_possible) > 0.0) ) {
    /*    (lp_best(ss->clause_set_possible) > lp_best(ss->clause_set_final))) { */
    loopcount++;
    if (DB(dbsearch)) {
      printf ("\n\n====Start hypothesis search loop\n");
      printf ("====count %d clause_set_possible size %d lpbest csp %3.3f\n\n", 
	      loopcount, 
	      lp_size(ss->clause_set_possible), 
	      lp_best(ss->clause_set_possible));
    }
    /*  printf("\n+++ Clause Set Possible List\n");
	lp_show(ss->clause_set_possible,t,i);*/
    v1 = lp_best(ss->clause_set_possible);
    lp_head(ss->clause_set_possible,
	    &p,&oldbits,&lpos,&last_bay);
   
    tab_cover(t,lpos,&pc1,&nc1,&tc1);
    theta = cal_theta(t,tc1);
    r = best_bay(t, oldbits, pc1, nc1, theta);
    
    
    if (r > lp_best(ss->clause_set_final)) {
      
      /*    lp_goodsmall(ss->clause_set_possible,lp_worse(ss->clause_set_final),
	    &p,&oldbits,&lpos,&last_bay);*/
      /*  printf("\n+++ Clause Set Possible List\n");
	  lp_show(ss->clause_set_possible,t,i);*/
     
      if (DB(dbsearchb)) {
	printf("\n\n\n====best so far %3.3f ",lp_best(ss->clause_set_final));
	show_lp (stdout, lp_bestlp(ss->clause_set_final), i);
	printf("====extending lp %d h: %3.3f last_bay:%3.3f poss_size:%3d ",p,v1,last_bay,
	       lp_size(ss->clause_set_possible));
	show_lp (stdout, p, i);
	printf ("\n\n");
      }
      for (j=0;j<lp_size(ss->clause_best);j++) {
	tlu = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lp;
	if (/*last_bay <= lp_bits(ss->clause_best,j) &&*/ 
	    /*TRUE*/
	   !lst_member(tlu,p)) {
	  
	  litp = fl_get(t->lit_free_list);      /* add to clause set possible */
	  lit = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lit;
	  
	  tab_cover(t,lpos,&pcold,&ncold,&tcold);
	 
	  bvs_or(t->lit_cover_p,lit,lpos,litp);
	  bvs_or(t->lit_cover_n,lit,lpos,litp);
	  bvs_or(t->lit_cover_t,lit,lpos,litp);
	  cal_cover(t,litp);
	  tab_cover(t,litp,&pc,&nc,&tc);
	  pn = lst_new(lst_size(p)+1);
	  lst_copy(p,pn);
	  lst_append(pn,tlu);
	  
	  pn2 = lst_new(lst_size(p)+1);                     
	  lst_copy(p,pn2);
	  lst_append(pn2,tlu);
	  litp2 = tab_replicate(t, litp );

	  bits = prob_lp(pn,i);
	  theta = cal_theta(t,tc);
	  poss_r = (sqrt((float) lst_size(pn))) * gain_lp(bits,pc,nc,pcold,ncold);
	  r = best_bay(t, bits, pc, nc, theta);
	  if ((poss_r > 0.0) && (r > lp_best(ss->clause_set_final))) {
	    if (DB(dbsearchb)) {
	      printf("====considering ");	
	      printf("j: %3d gain %2.4f ",j,poss_r);
	      printf("lpbits:%3.3f b:%2.4f pc:%6.2f nc:%6.2f po:%6.2f no:%6.2f t:%2.2e ",
		     lp_bits(ss->clause_best,j),
		     bits,
		     pc,
		     nc,
		     pcold,
		     ncold,
		     theta); 
	      show_c (stdout,tlu, i);
	      printf ("\n");
	    }
	    lp_add(ss->clause_set_possible,pn,poss_r,bits,litp,theta,t,lp_bits(ss->clause_best,j));
	  } else {
	    lst_free(pn);
	    fl_free(t->lit_free_list,litp);
	  }
	  
	  cal_cover(t,litp2);
	  tab_cover(t,litp2,&pc,&nc,&tc);
	  theta = cal_theta(t,tc);
	  r = bay(t, bits, pc, nc, theta);
	  if (!recursive) {
	    lp_add(ss->clause_set_final,pn2,r,bits,litp2,theta,t,0.0);
	  } else {
	    if(lp_better_then_worse(ss->clause_set_final,r)) {
	      if (DB(dbsearchb)) {
		printf ("\n==== Consider for resolution. %6.2f %6.2f %2.4e > %2.4e => ",
			pc,nc,r,lp_worse(ss->clause_set_final));
	      }
	      
	      cover(i,pn2,&steps,litp2);
	      
	      tab_cover(t,litp2,&pc,&nc,&tc);
	      theta = cal_theta(t,tc);
	      r = bay(t, bits, pc, nc, theta);
	      if (DB(dbsearchb)) {
		if (lp_better_then_worse(ss->clause_set_final,r)) {
		  printf(" Is better. %6.2f %6.2f %2.4e\n",pc,nc,r);
		} else {
		  printf(" Not better.%6.2f %6.2f %2.4e\n",pc,nc,r);
		}
		show_lp (stdout,pn2,i);
		printf("\n");
	      }
	      
	      lp_add(ss->clause_set_final,pn2,r,bits,litp2,theta,t,0.0);
	    } else {
	      lst_free(pn2);
	      fl_free(t->lit_free_list,litp2);
	    }
	    
	 } 
	
	}
      }
    } else {
      if (DB(dbsearch)) {
	printf("\n+++ Clause Set Final List\n");
	lp_show(ss->clause_set_final,t,i);
      }
      if (DB(dbsearchb)) {
	printf("\n\n\n====best so far %3.3f ",lp_best(ss->clause_set_final));
	show_lp (stdout, lp_bestlp(ss->clause_set_final), i);
	printf("====skip extending extending lp %d h: %3.3f last_bay:%3.3f poss_size:%3d pc1 %6.2f nc1 %6.2f\n",p,v1,last_bay,
	       lp_size(ss->clause_set_possible),pc1,nc1);
printf("r %6.2f < lpbest clause set final %6.2f\n", r , lp_best(ss->clause_set_final));
	show_lp (stdout, p, i);
	printf ("\n\n");
      }
    }
    
    lp_tree_remove(p,LEV_LP);
    fl_free(t->lit_free_list,lpos);
    
  }
  if (DB(dbsearch)) {
    printf ("\n\n====Completed hypothesis search loop\n");
    printf ("====count %d clause_set_possible size %d ", 
	    loopcount, 
	    lp_size(ss->clause_set_possible));
    if (lp_size(ss->clause_set_possible)>0) { 
      printf("lpbest csp %3.3f\n\n",
	    lp_best(ss->clause_set_possible));
    } else {
      printf("lpbest no value\n\n");
    }
  }

  
  if (DB(dbsearchb)) {
    printf("\n+++ Clause Set Possible List\n");
    lp_show(ss->clause_set_possible,t,i);
    
    printf("\n+++ Clause Set Final List\n");
    lp_show(ss->clause_set_final,t,i);
  }
  if (DB(dbstage)) printf("===loopcount %d\n", loopcount);
}


void
join_programs(search *ss,tab *t,info *i)
{
  int j;
  int litp, lit, lpos, lpostemp;
  float pc, nc, tc;
  float pc1, nc1, tc1;
  float pc2, nc2, tc2;
  lst pn, pn2, tlu;
  float bits, r,  theta;


  lpos = fl_get(t->lit_free_list);      /* add to clause set possible */
  lpostemp = fl_get(t->lit_free_list); 
  tab_clear(t,lpos);
  tab_clear(t,lpostemp);
  pn = lst_new(500);

  for (j=0;j<lp_size(ss->clause_best);j++) {
    if (DB(dbsearchb)) printf("====Start join for loop\n");
    tlu = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lp;
    lit = ((*(ss->clause_best)->data)[lst_look((ss->clause_best)->index,j)]).lit;
	
    litp = fl_get(t->lit_free_list);      /* add to clause set possible */
    bvs_or(t->lit_cover_p,lit,lpos,lpostemp);
    bvs_or(t->lit_cover_n,lit,lpos,lpostemp);
    bvs_or(t->lit_cover_t,lit,lpos,lpostemp);
    cal_cover(t,lpostemp);
    tab_cover(t,lpostemp,&pc,&nc,&tc);

    tab_cover(t,lit,&pc1,&nc1,&tc1);
    tab_cover(t,lpos,&pc2,&nc2,&tc2);

    if (((nc - nc2) < 1.0 && pc > pc2) ||
        ((pc-pc2)/(nc-nc2) > 2.0)) {

      tab_copy(t,lpostemp,lpos);
      lst_append(pn,tlu);
      
      if (DB(dbsearchb)) {
	printf("===The joint hyppothesis so far\n");
	show_lp (stdout,pn,i);
	printf("\n");    
      }
      
      
      litp = tab_replicate(t,lpos);
      pn2 = lst_replicate(pn);
      bits = 10.0; /*prob_lp(pn2,i);*/
      theta = cal_theta(t,tc);
      r = bay(t, bits, pc, nc, theta);
      r = j;
      lp_add(ss->clause_set_final,pn2,r,bits,litp,theta,t,0.0);
    } 
  }
  fl_free(t->lit_free_list,lpos); 
  fl_free(t->lit_free_list,lpostemp);
}


void
do_search(info *i,search *ss)
{
tab *t;

  t = i->the_tab;
	
  add_simple_clause_to_pos (ss,t,i);
  form_clauses (ss,t,i);
  invent_predicate(ss,t,i);
  add_clause_to_pos (ss,t,i);
  form_programs(ss,t,i); 
  /*  join_programs(ss,t,i); */
}




