/* utilities - memory, time, system stuff */

#include "main.h"

#include <stdlib.h>





/* returns a random number from 0 to n-1 */
int
my_random (int n)
{
  long r;
  r = rand();
  return (r % n);
}

float

my_random_float (float b)
{
  /*  long r;
  float f;
  
  r = rand();
  f = (r % 100001);
  return (f / 100000.0) * b; */
  return (b * (drand48()));
}

void
*memory (int size, char error_msg[])
{
void *x;

	mem_use++;
	x = malloc(size);
	if (x == NULL) {
		error(error_msg);
	}
	return(x);
}

void
mem_free (void *mem)
{
	mem_use--;
	free (mem);
}

void
error(char em[])
{
	fprintf(stderr,"%s\n -- Halting\n",em);
	exit(EXIT_FAILURE);
}

/* array stuff */

void
arr_cp(int num, int i[],int io, int o[], int oo)
{
int j;

	for (j=0;j<num;j++)
		o[oo + j] = i[io + j];
}

int
arr_cmp(int num, int a[], int ao, int b[], int bo)
{
int j;
	for (j=0;j<num;j++) {
		if (a[ao + j] != b[bo + j]) return (FALSE);
	}
	return (TRUE);
}

int
arr_cmp_lt(int num, int a[], int ao, int b[], int bo)
{
int j;
	for (j=0;j<num;j++) {
		if (a[ao + j] < b[bo + j])
			return (TRUE);
		else if (a[ao + j] > b[bo + j])
			return (FALSE);
	}
	return (FALSE);
}

fl
*fl_create(int size)
{
fl *f;
register int mem_req;
int i;

	mem_req = sizeof(int) * size;
	f = (fl *) memory (sizeof(fl),"create free list struct");
	f->next = 0;
	f->list = (int (*)[]) memory(mem_req, "creating data for free list");
	for (i=0;i<size-1;i++) (*f->list)[i] = i+1;
	(*f->list)[size-1] = -1;
	f->size = size;
	return(f);
}

void
fl_clear(fl *f)
{
  int i;

  f->next = 0;
  for (i=0;i<((f->size)-1);i++) (*f->list)[i] = i+1;
  (*f->list)[(f->size)-1] = -1;
}

void 
fl_dispose (fl *f)
{
  mem_free(f->list);
  mem_free(f);
}

int
fl_full(fl *f)
{
	return (f->next == -1);
}

int
fl_get(fl *f)
{
  int r;

  r = f->next;
  f->next = (*f->list)[r];
  (*f->list)[r] = -2;
  return (r);
}

int
fl_onfreelist(fl *f, int e) {
  int t;
  int i;

  i = 0;
  t = f->next;
  while (t != -1) {
    
    if (t == e) return TRUE;
    t = (*f->list)[t];
    if (i>f->size) error ("loop in free list");
    i++;
  }
  return FALSE;
}

void
fl_free(fl *f,int empty)
{
  /*if (fl_onfreelist(f,empty)) error("on free list");*/
  if ((*f->list)[empty] != -2) error("on free list");
  (*f->list)[empty] = f->next;
  f->next = empty;
}


/* lst stuff */

lst_str the_lst;

void
lst_create(int index_size, int data_size)
{
	the_lst.index = (index_entry (*)[]) memory (sizeof(index_entry) * index_size,"create index list");
	the_lst.index_free = fl_create(index_size);
}

lst
lst_new(int size)
{
	int idx;

	lst_use++;
	if (size < LST_DATA_SIZE) size = LST_DATA_SIZE;
	if (fl_full(the_lst.index_free)) error("index table full");
	idx = fl_get(the_lst.index_free);
	if (size <= LST_DATA_SIZE) {
		(*the_lst.index)[idx].start = (int (*)[]) (*the_lst.index)[idx].data;
	} else {
		(*the_lst.index)[idx].start = (int (*)[])
				memory (sizeof(int) * size, "mem for int list");
	}
	(*the_lst.index)[idx].num = 0;
	(*the_lst.index)[idx].max = size;
	return idx;
}

void
lst_grow(lst idx)
{
  int max, j;
  int (*ptr)[];

  max = (*the_lst.index)[idx].max * 2;
  ptr = (int (*)[]) memory (sizeof(int) * max, "mem for int list - grow");
  for (j=0;j<(*the_lst.index)[idx].num;j++) {
    (*ptr)[j] = (*(*the_lst.index)[idx].start)[j];
  }
  if ((*the_lst.index)[idx].max > LST_DATA_SIZE)
      mem_free((*the_lst.index)[idx].start);
  (*the_lst.index)[idx].start = ptr;
  (*the_lst.index)[idx].max = max;
}


void
lst_insert(lst idx, int d, int i)
{
	if (i >= (*the_lst.index)[idx].max) error("expand lst - inserting");
	if (i >= (*the_lst.index)[idx].num) (*the_lst.index)[idx].num = i + 1;
	(*(*the_lst.index)[idx].start)[i] = d;
}

void
lst_fill(lst idx, int num, int d)
{
  int j;

  while (num > (*the_lst.index)[idx].max) lst_grow(idx);
  (*the_lst.index)[idx].num = num;
  for(j=0;j<num;j++) 
    (*(*the_lst.index)[idx].start)[j] = d;
}

int
lst_append(lst idx, int d)
{
	int i;
        index_entry *iep;

        iep = &((*the_lst.index)[idx]);
	i = (*iep).num;
	if (i >= (*iep).max) lst_grow(idx);
	(*iep).num++;
	(*((*iep).start))[i] = d;
	return i;
}


int
lst_look(lst idx, int i)
{
	return (*(*the_lst.index)[idx].start)[i];
}

int
lst_size(lst idx)
{
	return (*the_lst.index)[idx].num;
}

void
lst_free(lst idx)
{
	lst_use--;
	if ((idx < 0) || (idx >= lst_msize)) error ("index overflow");
	if ((*the_lst.index)[idx].max > LST_DATA_SIZE)
		mem_free((*the_lst.index)[idx].start);
	fl_free(the_lst.index_free,idx);

}

int
lst_cmp(lst l1, lst l2)
{
	if ((*the_lst.index)[l1].num != (*the_lst.index)[l2].num)
		return FALSE;
	return arr_cmp ((*the_lst.index)[l1].num,
					*(*the_lst.index)[l1].start,
					0,
					*(*the_lst.index)[l2].start,
					0);
}

void
lst_shift(lst idx, int d, int i)
{
	int j;


	if (((*the_lst.index)[idx].num >= (*the_lst.index)[idx].max) ||
		(i >= (*the_lst.index)[idx].max))
		error("expand lst - shifting");
	if (i >= (*the_lst.index)[idx].num) {
		(*the_lst.index)[idx].num = i + 1;
		(*(*the_lst.index)[idx].start)[i] = d;
	} else {
		for(j=(*the_lst.index)[idx].num; j > i; j--)
			(*(*the_lst.index)[idx].start)[j] =
						(*(*the_lst.index)[idx].start)[j-1];
		(*(*the_lst.index)[idx].start)[i] = d;
		(*the_lst.index)[idx].num++;
	}
}

int
lst_pop(lst idx)
{
	int num;

	num = (*the_lst.index)[idx].num;
	if (num == 0) error ("popping empty");
	(*the_lst.index)[idx].num = num-1;
	return (*(*the_lst.index)[idx].start)[num-1];
}

int
lst_find(lst idx, int d)
{
	int j;
	int num;

	num = (*the_lst.index)[idx].num;
	for (j=0;j<num;j++)
		if (d == (*(*the_lst.index)[idx].start)[j]) return (j);
	return (-1);
}

int 
lstlst_find(lst x,lst xs)
{
   int k, size;
   
   size = lst_size(xs);
   for (k=0;k<size;k++) {
      if (lst_cmp(lst_look(xs,k),x)) return k;
   }
   return (-1);
}


int 
lstlst_member(lst x,lst xs)
{
   int k, size;
   
   size = lst_size(xs);
   for (k=0;k<size;k++) {
      if (lst_cmp(lst_look(xs,k),x)) return TRUE;
   }
   return FALSE;
}


int
lst_member(int d, lst idx)
{
	return (lst_find(idx,d) != -1);
}

void
lst_show(lst idx)
{
	int j;
	int num;

	num = (*the_lst.index)[idx].num;
	for (j=0;j<num;j++)
		printf("%d ",(*(*the_lst.index)[idx].start)[j]);

}

void
bv_to_lst(bvs *s, lst res, int lit)
{
int j;

	lst_clear(res);
	for (j=0;j<s->vector_size;j++) {
		if (bvs_get(s,lit,j)) {
			lst_append(res,j);
		}
	}

}

void
lst_clear(lst idx)
{
	(*the_lst.index)[idx].num = 0;
}

int
lst_head(lst idx)
{
	int num, res, j;

	num = (*the_lst.index)[idx].num;
	if (num == 0) error ("lst head empty");
	res = (*(*the_lst.index)[idx].start)[0];
	(*the_lst.index)[idx].num = num-1;
	for(j=0;j<num-1;j++)
		(*(*the_lst.index)[idx].start)[j] =
			(*(*the_lst.index)[idx].start)[j+1];
	return (res);
}
int
lst_del(lst idx,int i)
{
	int num, res, j;

	num = (*the_lst.index)[idx].num;
	if (num == 0) error ("lst empty");
	if (num <= i) error ("deleteing off end of the list");
	res = (*(*the_lst.index)[idx].start)[0];
	(*the_lst.index)[idx].num = num-1;
	for(j=i;j<num-1;j++)
		(*(*the_lst.index)[idx].start)[j] =
			(*(*the_lst.index)[idx].start)[j+1];
	return (res);
}

void
lst_copy (lst s, lst d)
{
  int j, num;

  lst_clear(d);
  num  = lst_size(s);
  for (j=0;j<num;j++)
    lst_append(d,lst_look(s,j));
	
}

int
lst_sum (lst d)
{
  int j, sum, num;

  sum = 0;
  num  = lst_size(d);
  for (j=0;j<num;j++)
    sum += lst_look(d,j);
  return sum;
}



int 
lst_subset(lst ss, lst s)
{
  int ai,bi,a,b, size_ss, size_s;

  ai=0;
  size_ss = lst_size(ss);
  size_s = lst_size(s);

  for (bi=0;bi<size_s;bi++) {
    if (ai == size_ss) return TRUE;
    a = lst_look(ss,ai);
    b = lst_look(s,bi);
    if (a == b) {
      ai++;
    } else if (a < b) return FALSE;
  }
  return (ai == size_ss);
}

void
lst_merge(lst s1, lst s2, lst d)
{
  int a,b;
  int size_s1, size_s2;
  int lls1a, lls2b;

  lst_clear(d);
  a = 0;
  b = 0;
  size_s1 = lst_size(s1);
  size_s2 = lst_size(s2);
  while ((a < size_s1) || (b < size_s2)) {
    lls1a = lst_look(s1,a);
    lls2b = lst_look(s2,b);
    if (a == size_s1) {
      lst_append(d,lls2b);
      b++;
    } else if (b == size_s2) {
      lst_append(d,lls1a);
      a++;
    } else {
      if (lls1a == lls2b) {
	lst_append(d,lls1a);
	a++;
	b++;
      } else if (lls1a < lls2b) {
	lst_append(d,lls1a);
	a++;
      } else {
       lst_append(d,lls2b);
       b++;
     }
    }
  }
}

void
lst_2_arr(lst l,int a[])
{
  int k;

  for (k=0;k<lst_size(l);k++) {
    a[k] = lst_look(l,k);
  }
}

lst
lst_replicate(lst x) 
{
  lst y;

  y = lst_new(lst_size(x));
  lst_copy(x,y);
  return (y);
}


void
arr_2_lst(int a[], int num, lst l)
{
  int k;

  lst_clear(l);
  for (k=0;k<num;k++)
    lst_append(l,a[k]);
}

int 
lst_max (lst l, int a)
{
  int k,t,m,s;

  m = a;
  s = lst_size(l);
  for (k=0;k<s;k++) {
    if ((t =lst_look(l,k))>m) m = t;
  }
  return m;
}

int
max(int a ,int b)
{
  if (a>b) return a;
  else return b;
}


/* float int fudge */ 

float
mir_float(int i) {
  union floatint fi;
  fi.i = i;
  return fi.f;
}

int
mir_int(float f) {
  union floatint fi;
  fi.f = f;
  return fi.i;
}

void
mir_check() {
  if (sizeof(int) != sizeof(float)) error("int float fudge will not work");
}




/* random list */

lst rlist;

void 
init_rlist ()
{
  rlist = lst_new(100);
}

int 
look_rlist (int index)
{
  while (index >= lst_size(rlist)) { /* extend the rlist */
    lst_append(rlist,my_random(30013));
  }
  return (lst_look(rlist,index));
}
